#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vector3D.h>

float randomNumber(float min, float max)
{
  float alpha;

  alpha= (float)(rand())/(float)(RAND_MAX); 

  return ( min + alpha*(max-min) );
}

Vector2D::Vector2D()
{ 
  x= y= 0;
}

Vector2D::Vector2D(const Vector2D &v)
{ 
  x= v.x;
  y= v.y;
}


Vector2D::Vector2D(float x0, float y0)
{ 
  x= x0;
  y= y0;
}

void Vector2D::set(float x0, float y0)
{ 
  x= x0;
  y= y0;
}

Vector2D Vector2D::operator + () const
{
  Vector2D result;

  result.x= x;
  result.y= y;

  return result;
}

Vector2D Vector2D::operator - () const
{
  Vector2D result;

  result.x= -x;
  result.y= -y;

  return result;
}

// Vector2D Vector2D::operator + (const Vector2D &v)
// {
//   Vector2D result;
// 
//   result.x= x + v.x;
//   result.y= y + v.y;
// 
//   return result;
// }

// Vector2D Vector2D::operator - (const Vector2D &v)
// {
//   Vector2D result;
// 
//   result.x= x - v.x;
//   result.y= y - v.y;
// 
//   return result;
// }

Vector2D operator + (const Vector2D &v1, const Vector2D &v2)
{
  Vector2D result;

  result.x= v1.x + v2.x;
  result.y= v1.y + v2.y;

  return result;
}

Vector2D operator - (const Vector2D &v1, const Vector2D &v2)
{
  Vector2D result;

  result.x= v1.x - v2.x;
  result.y= v1.y - v2.y;

  return result;
}

void Vector2D::operator += (const Vector2D& v)
{
   x+= v.x;
   y+= v.y;
}

void Vector2D::operator -= (const Vector2D& v)
{
   x-= v.x;
   y-= v.y;
}

void Vector2D::operator *= (float s)
{
   x*= s;
   y*= s;
}

void Vector2D::operator /= (float s)
{
   x/= s;
   y/= s;
}

Vector2D Vector2D::operator * (float a)
{
  Vector2D result;

  result.x= a*x;
  result.y= a*y;

  return result;
}

Vector2D operator * (float a, Vector2D v) 
{
  Vector2D result;

  result.x= a*v.x;
  result.y= a*v.y;

  return result;
}

Vector2D Vector2D::operator / (float a)
{
  Vector2D result;

  result.x= x/a;
  result.y= y/a;

  return result;
}

float Vector2D::operator () (unsigned int i) const
{
  switch(i)
  {
    case 1:
      return x;
    break;
    case 2:
      return y;
    break;
    default:
      printf("Vector2D::operator (): index exceeds vector dimensions.\n");
      return 0;
    break;
  }
}

float& Vector2D::operator () (unsigned int i)
{
  switch(i)
  {
    case 1:
      return x;
    break;
    case 2:
      return y;
    break;
    default:
      printf("Vector2D::operator (): index exceeds vector dimensions.\n");
      return x;
    break;
  }
}

float Vector2D::operator [] (unsigned int i) const
{
  switch(i)
  {
    case 0:
      return x;
    break;
    case 1:
      return y;
    break;
    default:
      printf("Vector2D::operator []: index exceeds vector dimensions.\n");
      return 0;
    break;
  }
}

float& Vector2D::operator [] (unsigned int i)
{
  switch(i)
  {
    case 0:
      return x;
    break;
    case 1:
      return y;
    break;
    default:
      printf("Vector2D::operator []: index exceeds vector dimensions.\n");
      return x;
    break;
  }
}

bool Vector2D::operator == (const Vector2D& v) const
{
   float d;

   d= sqrt( pow(x-v.x, 2) + pow(y-v.y, 2) );

   if(d<EPSILON)
     return true;
   else 
     return false;
}


float Vector2D::norm()
{
  return sqrt( x*x + y*y );
}

int Vector2D::normalize()
{
  float norm;
  norm= sqrt( x*x + y*y );

  if(norm>EPSILON)
  {
    x/= norm;
    y/= norm;
    return 1;
  }
  else
  {
    //printf("%s: %d: Vector2D::normalize(): the vector to normalize has a null norm.\n", __FILE__, __LINE__);
    x= 0.0;
    y= 0.0;
    return 0;
  }

}

void Vector2D::print(const char * name)
{
  printf("%s= \n", name);
  printf("    [ %f ]\n", x);
  printf("    [ %f ]\n", y);
}


//! Computes an (arbitrary) vector that is orthogonal to the calling vector.
Vector2D Vector2D::orthogonalVector()
{
   Vector2D result;
   #ifdef DEBUG
   if( this->norm()<EPSILON )
   {
      printf("%s: %d: Vector2D::orthogonalVector(): bad input (zero norm vector).\n",__FILE__,__LINE__);
      return result;
   }
   #endif

   if( fabs(y) <= EPSILON )
   {
     result.x= 0;
     result.y= 1; 
     return result;
   }
   else
   {
     result.x= 1;
     result.y= -x/y; 
     result.normalize();
     return result;
   }
}

float dotProduct(const Vector2D &v1, const Vector2D &v2)
{
  return (v1.x*v2.x + v1.y*v2.y);
}

float norm(const Vector2D &v)
{
  return sqrt(v.x*v.x + v.y*v.y);
}

Vector3D::Vector3D()
{ 
  x= y= z= 0;
}

Vector3D::Vector3D(float x0, float y0, float z0)
{ 
  x= x0;
  y= y0;
  z= z0; 
}

Vector3D::Vector3D(const Vector3D &v)
{ 
  x= v.x;
  y= v.y;
  z= v.z;
}

void Vector3D::set(float x0, float y0, float z0)
{ 
  x= x0;
  y= y0;
  z= z0; 
}

Vector3D Vector3D::operator + () const
{
  return Vector3D(x, y, z);
}

Vector3D Vector3D::operator - () const
{
  return Vector3D(-x, -y, -z);
}

// Vector3D Vector3D::operator + (const Vector3D &v)
// {
//   Vector3D result;
// 
//   result.x= x + v.x;
//   result.y= y + v.y;
//   result.z= z + v.z;
// 
//   return result;
// }

// Vector3D Vector3D::operator - (const Vector3D &v)
// {
//   Vector3D result;
// 
//   result.x= x - v.x;
//   result.y= y - v.y;
//   result.z= z - v.z;
// 
//   return result;
// }

Vector3D operator + (const Vector3D &v1, const Vector3D &v2)
{
  Vector3D result;

  result.x= v1.x + v2.x;
  result.y= v1.y + v2.y;
  result.z= v1.z + v2.z;

  return result;
}

Vector3D operator - (const Vector3D &v1, const Vector3D &v2)
{
  Vector3D result;

  result.x= v1.x - v2.x;
  result.y= v1.y - v2.y;
  result.z= v1.z - v2.z;

  return result;
}

void Vector3D::operator += (const Vector3D& v)
{
   x+= v.x;
   y+= v.y;
   z+= v.z;
}

void Vector3D::operator -= (const Vector3D& v)
{
   x-= v.x;
   y-= v.y;
   z-= v.z;
}

void Vector3D::operator *= (float s)
{
   x*= s;
   y*= s;
   z*= s;
}

void Vector3D::operator /= (float s)
{
   x/= s;
   y/= s;
   z/= s;
}

Vector3D Vector3D::operator * (float a)
{
  Vector3D result;

  result.x= a*x;
  result.y= a*y;
  result.z= a*z;

  return result;
}

Vector3D operator * (float a , Vector3D v) 
{
  Vector3D result;

  result.x= a*v.x;
  result.y= a*v.y;
  result.z= a*v.z;

  return result;
}

Vector3D operator * (const HTMatrix &M, Vector3D v) 
{
  Vector3D result;

  result.x= M.m11*v.x + M.m12*v.y + M.m13*v.z + M.m14;
  result.y= M.m21*v.x + M.m22*v.y + M.m23*v.z + M.m24;
  result.z= M.m31*v.x + M.m32*v.y + M.m33*v.z + M.m34;

  return result;
}

Vector3D Vector3D::operator / (float a)
{
  Vector3D result;

  result.x= x/a;
  result.y= y/a;
  result.z= z/a;

  return result;
}

float Vector3D::operator () (unsigned int i) const
{
  switch(i)
  {
    case 1:
      return x;
    break;
    case 2:
      return y;
    break;
    case 3:
      return z;
    break;
    default:
      printf("Vector3D::operator (): index exceeds vector dimensions.\n");
      return 0;
    break;
  }
}

float& Vector3D::operator () (unsigned int i)
{
  switch(i)
  {
    case 1:
      return x;
    break;
    case 2:
      return y;
    break;
    case 3:
      return z;
    break;
    default:
      printf("Vector3D::operator (): index exceeds vector dimensions.\n");
      return x;
    break;
  }
}

float Vector3D::operator [] (unsigned int i) const
{
  switch(i)
  {
    case 0:
      return x;
    break;
    case 1:
      return y;
    break;
    case 2:
      return z;
    break;
    default:
      printf("Vector3D::operator []: index exceeds vector dimensions.\n");
      return 0;
    break;
  }
}

float& Vector3D::operator [] (unsigned int i)
{
  switch(i)
  {
    case 0:
      return x;
    break;
    case 1:
      return y;
    break;
    case 2:
      return z;
    break;
    default:
      printf("Vector3D::operator []: index exceeds vector dimensions.\n");
      return x;
    break;
  }
}

bool Vector3D::operator == (const Vector3D& v) const
{
   float d;

   d= sqrt( pow(x-v.x, 2) + pow(y-v.y, 2) + pow(z-v.z, 2) );

   if(d<EPSILON)
     return true;
   else 
     return false;
}


float Vector3D::norm()
{
  return sqrt( x*x + y*y + z*z);
}

int Vector3D::normalize()
{
  float norm;
  norm= sqrt( x*x + y*y + z*z);

  if(norm>EPSILON)
  {
    x/= norm;
    y/= norm;
    z/= norm;
    return 1;
  }
  else
  {
    //printf("%s: %d: Vector3D::normalize(): the vector to normalize has a null norm.\n", __FILE__, __LINE__);
    x= 0.0;
    y= 0.0;
    z= 0.0;
    return 0;
  }

}

void Vector3D::print(const char * name)
{
  printf("%s= \n", name);
  printf("    [ %f ]\n", x);
  printf("    [ %f ]\n", y);
  printf("    [ %f ]\n", z);
}

//! Computes an (arbitrary) vector that is orthogonal to the calling vector.
Vector3D Vector3D::orthogonalVector()
{
   Vector3D result;
   #ifdef DEBUG
   if( this->norm()<EPSILON )
   {
     printf("%s: %d: Vector3D::orthogonalVector(): bad input (zero norm vector).\n",__FILE__,__LINE__);
     return result;
   }
   #endif

   if( fabs(z) <= EPSILON )
   {
     result.x= 0;
     result.y= 0; 
     result.z= 1;
     return result;
   }
   else
   {
     result.x= 0;
     result.y= 1; 
     result.z= -y/z;
     result.normalize();
     return result;
   }
}

float dotProduct(const Vector3D &v1, const Vector3D &v2)
{
  return (v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);
}

float norm(const Vector3D &v)
{
  return sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
}

Vector3D crossProduct(const Vector3D &v1, const Vector3D &v2)
{
  Vector3D result;

  result.x= v1.y*v2.z - v1.z*v2.y;
  result.y= v1.z*v2.x - v1.x*v2.z;
  result.z= v1.x*v2.y - v1.y*v2.x;

  return result;
}


float HTMatrix::operator () (unsigned int i, unsigned int j) const
{
  if( (i<1) || (i>3) || (j<1) || (j>4) )
  {
     printf("HTMatrix3D::operator (): index exceeds matrix dimensions.\n");
     return 0;
  }

  if(i==1)
  {
    if(j==1)
      return m11;
    if(j==2)
      return m12;
    if(j==3)
      return m13;
    if(j==4)
      return m14;
  }

  if(i==2)
  {
    if(j==1)
      return m21;
    if(j==2)
      return m22;
    if(j==3)
      return m23;
    if(j==4)
      return m24;
  }

  if(i==3)
  {
    if(j==1)
      return m31;
    if(j==2)
      return m32;
    if(j==3)
      return m33;
    if(j==4)
      return m34;
  }

  if(i==4)
  {
    if(j==1)
      return 0;
    if(j==2)
      return 0;
    if(j==3)
      return 0;
    if(j==4)
      return 1;
  }

  return 0;
}

float& HTMatrix::operator () (unsigned int i, unsigned int j)
{
  if( (i<1) || (i>3) || (j<1) || (j>4) )
  {
     printf("HTMatrix3D::operator (): index exceeds matrix dimensions.\n");
     return m11;
  }

  if(i==1)
  {
    if(j==1)
      return m11;
    if(j==2)
      return m12;
    if(j==3)
      return m13;
    if(j==4)
      return m14;
  }

  if(i==2)
  {
    if(j==1)
      return m21;
    if(j==2)
      return m22;
    if(j==3)
      return m23;
    if(j==4)
      return m24;
  }

  if(i==3)
  {
    if(j==1)
      return m31;
    if(j==2)
      return m32;
    if(j==3)
      return m33;
    if(j==4)
      return m34;
  }

  return m11;
}

HTMatrix HTMatrix::operator * (const HTMatrix &mat)
{
  HTMatrix result;

  result.m11= m11*mat.m11 + m12*mat.m21 + m13*mat.m31;
  result.m12= m11*mat.m12 + m12*mat.m22 + m13*mat.m32;
  result.m13= m11*mat.m13 + m12*mat.m23 + m13*mat.m33;
  result.m14= m11*mat.m14 + m12*mat.m24 + m13*mat.m34 + m14;

  result.m21= m21*mat.m11 + m22*mat.m21 + m23*mat.m31;
  result.m22= m21*mat.m12 + m22*mat.m22 + m23*mat.m32;
  result.m23= m21*mat.m13 + m22*mat.m23 + m23*mat.m33;
  result.m24= m21*mat.m14 + m22*mat.m24 + m23*mat.m34 + m24;

  result.m31= m31*mat.m11 + m32*mat.m21 + m33*mat.m31;
  result.m32= m31*mat.m12 + m32*mat.m22 + m33*mat.m32;
  result.m33= m31*mat.m13 + m32*mat.m23 + m33*mat.m33;
  result.m34= m31*mat.m14 + m32*mat.m24 + m33*mat.m34 + m34;

  return result;
}

void HTMatrix::operator *= (const HTMatrix &mat)
{
  (*this)=(*this)*mat;
}


void HTMatrix::setIdentity()
{
   m11= m22= m33= 1.0;
   m12= m13= m14= m21= m23= m24= m31= m32= m34= 0.0;
}

//! Sets the translation part of a HTMatrix.
void HTMatrix::setTranslation(float x, float y, float z)
{
   m14= x;
   m24= y;
   m34= z;
}

//! Sets the rotation part of a HTMatrix from an angle and a rotation axis.
//! \param theta angle of the rotation (given in radians)
//! \param ax X coordinate of the rotation axis
//! \param ay Y coordinate of the rotation axis
//! \param az Z coordinate of the rotation axis
void HTMatrix::setRotationFromAngleAndAxis(float theta, float ax, float ay, float az)
{
  float x, y, z, norm;
  float c, s, C, xs, ys, zs, xC, yC, zC, xyC, yzC, zxC;

  norm= sqrt(ax*ax + ay*ay + az*az);
  x= ax/norm;
  y= ay/norm;
  z= az/norm;

  c = cos(theta);
  s = sin(theta);
  C = 1-c;
  xs = x*s;
  ys = y*s;
  zs = z*s;
  xC = x*C; 
  yC = y*C; 
  zC = z*C;

  xyC = x*yC;
  yzC = y*zC;
  zxC = z*xC;


  m11= x*xC+c;     m12= xyC-zs;     m13= zxC+ys;
  m21= xyC+zs;     m22= y*yC+c;     m23= yzC-xs;
  m31= zxC-ys;     m32= yzC+xs;     m33= z*zC+c;
}



//! Sets the rotation part of the homogeneous matrix from yaw, pitch and roll angles given in radians (Tait-Bryan angles).
//! Performs a rotation around initial Z axis, followed by a rotation around the new Y axis, 
//! followed by a rotation around the new X axis.
//! \param yaw angle value of the rotation around X 
//! \param pitch angle value of the rotation around Y 
//! \param roll angle value of the rotation around Z 
void HTMatrix::setRotationFromYPRangles(float yaw, float pitch, float roll)
{
  float ax, ay, az;
  float cx, sx, cy, sy, cz, sz, czsy, szsy;

  az= yaw;
  ay= pitch;
  ax= roll;

  sx = sin(ax);
  cx = cos(ax);
  sy = sin(ay);
  cy = cos(ay);
  sz = sin(az);
  cz = cos(az);

  czsy = cz*sy;
  szsy = sz*sy;


  m11 = cz*cy;     m12 = czsy*sx - sz*cx;    m13 = czsy*cx + sz*sx;
  m21 = sz*cy;     m22 = szsy*sx + cz*cx;    m23 = szsy*cx - cz*sx;
  m31 = -sy;       m32 = cy*sx;              m33 = cy*cx;
}


//! Creates a frame matrix from a point and an axis (one among the infinity of possible frames).
//! \param c center of the frame
//! \param x x coordinate of the X axis of the frame
//! \param y y coordinate of the X axis of the frame
//! \param z z coordinate of the X axis of the frame
//! \return 1 in case of success, 0 otherwise.
int HTMatrix::setFromPointAndAxis(const Vector3D &c, float x, float y, float z)
{
   Vector3D axis(x, y, z), v, w;

   m14= c.x;
   m24= c.y;
   m34= c.z;

   if(axis.normalize()==0)
   {
    printf("%s: %d: HTMatrix::setFromPointAndAxis(): the input axis is null.\n", __FILE__, __LINE__);
    this->setIdentity();
    return 0;
   }

   v= axis.orthogonalVector();
   v.normalize();

   w= crossProduct(axis, v);
   w.normalize();   

   m11= axis.x;   m12= v.x;   m13= w.x;
   m21= axis.y;   m22= v.y;   m23= w.y;
   m31= axis.z;   m32= v.z;   m33= w.z;

   return 1;
}

//! Creates a transformation matrix from Denavit-Hartenberg parameters. Angles are given in radians.
//! \param a link length (translation along x) 
//! \param alpha link twist (rotation around x) (given in radians)
//! \param d link offset (translation along z)
//! \param theta joint angle (rotation around z) (given in radians)
//! \return 1 in case of success, 0 otherwise.
int HTMatrix::setFromDHparameters(float a, float alpha, float d, float theta)
{
  m11=  cos(theta);    m12= -sin(theta)*cos(alpha);   m13= sin(theta)*sin(alpha);     m14= a*cos(theta);
  m21=  sin(theta);    m22= cos(theta)*cos(alpha);    m23= -cos(theta)*sin(alpha);    m24= a*sin(theta);
  m31=           0;    m32= sin(alpha);               m33= cos(alpha);                m34= d;

  return 1;
}

Vector3D HTMatrix::operator * (const Vector3D &v)
{
   Vector3D result;

   result.x= m11*v.x + m12*v.y + m13*v.z + m14;
   result.y= m21*v.x + m22*v.y + m23*v.z + m24;
   result.z= m31*v.x + m32*v.y + m33*v.z + m34;

   return result;
}

//! Returns the product of the HTMatrix's rotation part by the vector v.
Vector3D rotationMult(const HTMatrix &mat, const Vector3D &v)
{
   Vector3D result;

   result.x= mat.m11*v.x + mat.m12*v.y + mat.m13*v.z;
   result.y= mat.m21*v.x + mat.m22*v.y + mat.m23*v.z;
   result.z= mat.m31*v.x + mat.m32*v.y + mat.m33*v.z ;

   return result;
}



HTMatrix inverse(const HTMatrix &mat)
{
  HTMatrix result;

  result.m11= mat.m11;
  result.m12= mat.m21;
  result.m13= mat.m31;

  result.m21= mat.m12;
  result.m22= mat.m22;
  result.m23= mat.m32;

  result.m31= mat.m13;
  result.m32= mat.m23;
  result.m33= mat.m33;

  result.m14= -( mat.m11*mat.m14 + mat.m21*mat.m24 + mat.m31*mat.m34 );
  result.m24= -( mat.m12*mat.m14 + mat.m22*mat.m24 + mat.m32*mat.m34 );      
  result.m34= -( mat.m13*mat.m14 + mat.m23*mat.m24 + mat.m33*mat.m34 );      

  return result;
}

void HTMatrix::invert()
{
  HTMatrix inv= inverse(*this);

  m11= inv.m11;  m12= inv.m12;  m13= inv.m13;  m14= inv.m14;
  m21= inv.m21;  m22= inv.m22;  m23= inv.m23;  m24= inv.m24;
  m31= inv.m31;  m32= inv.m32;  m33= inv.m33;  m34= inv.m34;
}

void HTMatrix::print(std::string name)
{
  printf("%s= [\n", name.c_str());
  printf("\t %f %f %f %f\n", m11, m12, m13, m14);
  printf("\t %f %f %f %f\n", m21, m22, m23, m24);
  printf("\t %f %f %f %f\n", m31, m32, m33, m34);
  printf("\t %f %f %f %f ]\n", 0.0,  0.0,  0.0,  1.0);
}


//! Returns the first column of the matrix as a 3D vector.
Vector3D HTMatrix::firstColumn()
{
   Vector3D result;
   result.x= m11;
   result.y= m21;
   result.z= m31;

   return result; 
}

//! Returns the second column of the matrix as a 3D vector.
Vector3D HTMatrix::secondColumn()
{
   Vector3D result;
   result.x= m12;
   result.y= m22;
   result.z= m32;

   return result; 
}

//! Returns the third column of the matrix as a 3D vector.
Vector3D HTMatrix::thirdColumn()
{
   Vector3D result;
   result.x= m13;
   result.y= m23;
   result.z= m33;

   return result; 
}

//! Returns the fourth column of the matrix as a 3D vector.
Vector3D HTMatrix::fourthColumn()
{
   Vector3D result;
   result.x= m14;
   result.y= m24;
   result.z= m34;

   return result; 
}


//! Sets the value of the first column of the matrix from a 3D vector.
void HTMatrix::setFirstColumn(Vector3D &v)
{
   m11= v.x;
   m21= v.y;
   m31= v.z;
}

//! Sets the value of the second column of the matrix from a 3D vector.
void HTMatrix::setSecondColumn(Vector3D &v)
{
   m12= v.x;
   m22= v.y;
   m32= v.z;
}

//! Sets the value of the third column of the matrix from a 3D vector.
void HTMatrix::setThirdColumn(Vector3D &v)
{
   m13= v.x;
   m23= v.y;
   m33= v.z;
}

//! Sets the value of the fourth column of the matrix from a 3D vector.
void HTMatrix::setFourthColumn(Vector3D &v)
{
   m14= v.x;
   m24= v.y;
   m34= v.z;
}

void HTMatrix::setFrom4x4Array(float mat[4][4])
{
  m11= mat[0][0];     m12= mat[0][1];     m13= mat[0][2];     m14= mat[0][3]; 
  m21= mat[1][0];     m22= mat[1][1];     m23= mat[1][2];     m24= mat[1][3]; 
  m31= mat[2][0];     m32= mat[2][1];     m33= mat[2][2];     m34= mat[2][3]; 
}

#if defined(ENABLE_DISPLAY)
//! Writes the content of the HTMatrix in an array with the format used by OpenGL (when calling a function
//! like glLoadMatrix or glMultMatrix).
void HTMatrix::convertToOpenGLFormat(GLfloat mat[16])
{
  mat[0]= m11;    mat[4]= m12;    mat[8]=  m13;    mat[12]= m14;
  mat[1]= m21;    mat[5]= m22;    mat[9]=  m23;    mat[13]= m24;
  mat[2]= m31;    mat[6]= m32;    mat[10]= m33;    mat[14]= m34;
  mat[3]=   0;    mat[7]=   0;    mat[11]=   0;    mat[15]=   1;
}
#endif


Matrix::Matrix()
{
  data= NULL;
  nbRows_= nbColumns_= 0;
}

Matrix::Matrix(unsigned int nbRows, unsigned int nbColumns)
{
  data= NULL;
  nbRows_= nbRows;
  nbColumns_= nbColumns;

  data= new float[nbRows_*nbColumns_];

  if(data==NULL)
  {
   printf("%s: %d: Matrix::Matrix(unsigned int, unsigned int): memory allocation failure.\n",__FILE__,__LINE__);
  }
  
  setToZero();
}


Matrix::Matrix(const Matrix &mat)
{
  unsigned i;

  resize(mat.nbRows(), mat.nbColumns());

  for(i=0; i<nbRows_*nbColumns_; ++i)
  {
    data[i]= mat.data[i];
  }
}

int Matrix::resize(unsigned int nbRows, unsigned int nbColumns)
{
  if(data!=NULL)
  {
    if( (nbRows_==nbRows) && (nbColumns_==nbColumns) )
    {
      return 1;
    }
    else
    {
      delete [] data;
      nbRows_= nbRows;
      nbColumns_= nbColumns;
      data= new float[nbRows_*nbColumns_];
      setToZero();
      return 1;
    }
  }
  else
  {
    nbRows_= nbRows;
    nbColumns_= nbColumns;
    data= new float[nbRows_*nbColumns_];
    setToZero();
    return 1;
  }

}


int Matrix::setToZero()
{
  unsigned int i;

  for(i=0; i<nbRows_*nbColumns_; ++i)
  {
    data[i]= 0.0;
  }

  return 1;
}


float Matrix::operator () (unsigned int i, unsigned int j) const
{
  if( (i<1) || (i>nbRows_) || (j<1) || (j>nbColumns_) || (nbRows_==0) || (nbColumns_==0) )
  {
     printf("%s: %d: Matrix::operator (): index exceeds matrix dimensions.\n",__FILE__,__LINE__);
     return 0;
  }

  return data[j-1 + (i-1)*nbColumns_];
}

float& Matrix::operator () (unsigned int i, unsigned int j)
{
  if( (i<1) || (i>nbRows_) || (j<1) || (j>nbColumns_) || (nbRows_==0) || (nbColumns_==0) )
  {
     printf("%s: %d: Matrix::operator (): index exceeds matrix dimensions.\n",__FILE__,__LINE__);
     return data[0];
  }

  return data[j-1 + (i-1)*nbColumns_];
}

Matrix & Matrix::operator = (const Matrix &mat)
{
  unsigned i;

  if(this==&mat)
  {  return *this;  }  

  resize(mat.nbRows(), mat.nbColumns());

  for(i=0; i<nbRows_*nbColumns_; ++i)
  {
    data[i]= mat.data[i];
  }

  return *this;
}

Matrix Matrix::operator * (const Matrix &mat)
{
   unsigned int i, j, k;
   float sum;
   Matrix result;

   if(nbColumns_ != mat.nbRows_)
   {
     printf("%s: %d: Matrix::operator *: inconsistent multiplication.\n",__FILE__,__LINE__);
     return result;
   }

   result= Matrix(nbRows_, mat.nbColumns_);

   
   for(i=1; i<=nbRows_; ++i)
   {
     for(j=1; j<=mat.nbColumns_; j++)
      {
        sum= 0;
	for(k=1; k<=nbColumns_; k++)
	{ 
          sum+= (*this)(i, k)*mat(k, j);
        }
	result(i, j)= sum;
      }
   }

   return result;
}


void Matrix::operator *= (const Matrix &mat)
{
  (*this)= (*this)*mat;
}

void Matrix::print(std::string name)
{
  unsigned int i, j;
  printf("%s=\n [\n", name.c_str());

  for(i=1; i<=nbRows_; ++i)
  {
    printf("   ");
    for(j=1; j<=nbColumns_; j++)
    {
      printf("%3.4f ", (*this)(i,j));
    }
    printf("\n");
  }
  printf(" ]\n");
}

int Matrix::setToIdentity(unsigned int size)
{
  unsigned int i;
  if(data!=NULL)
  {
    if( nbRows_!=size || nbColumns_!=size )
    {
      delete [] data;
      data= new float[size*size];
    }
  }
  else
  {
    data= new float[size*size];
  }

  nbRows_= size;
  nbColumns_= size;

  for(i=0; i<size*size; ++i)
  {
    data[i]= 0;
  }
  for(i=1; i<=size; ++i)
  {
    (*this)(i, i)= 1.0;
  }

  return 1;
}

int inverse(Matrix &mat, Matrix &result)
{
  float det;
  Matrix cofactor(3, 3); //for 3x3 inversion

  if( mat.nbRows()!=mat.nbColumns() )
  {
    printf("%s: %d: inverse(Matrix &, Matrix &): input matrix must be square.\n",__FILE__,__LINE__);
    return 0;
  }

  //for arbitrary-sized matrix inversion:
  int dim= mat.nbRows();
  int i, j, k, l, err;
  int nmax= dim+1, n2max= 2*nmax-1;
  float a[nmax][nmax], b[nmax][nmax];
  float max, pivot, coef;
  float t[nmax][n2max];


  result.resize(mat.nbRows(), mat.nbColumns());

  switch(mat.nbRows())
  {
    case 0:
    break;
    case 1:
       if( fabs(mat(1,1)) < EPSILON )
       { return 0; }

       result(1,1)= 1.0/mat(1,1);
    break;  
    case 2:
       det= mat(1,1)*mat(2,2) - mat(2,1)*mat(1,2);
       if( fabs(det) < EPSILON )
       { return 0; }

       result(1,1)=  mat(2,2)/det;
       result(1,2)= -mat(1,2)/det;
       result(2,1)= -mat(2,1)/det;
       result(2,2)=  mat(1,1)/det;
    break;
    case 3:
       cofactor(1,1)=    mat(2,2)*mat(3,3) - mat(3,2)*mat(2,3);
       cofactor(1,2)= -( mat(2,1)*mat(3,3) - mat(3,1)*mat(2,3) );
       cofactor(1,3)=    mat(2,1)*mat(3,2) - mat(3,1)*mat(2,2);

       cofactor(2,1)= -( mat(1,2)*mat(3,3) - mat(3,2)*mat(1,3) );
       cofactor(2,2)=    mat(1,1)*mat(3,3) - mat(3,1)*mat(1,3);
       cofactor(2,3)= -( mat(1,1)*mat(3,2) - mat(3,1)*mat(1,2) );

       cofactor(3,1)=    mat(1,2)*mat(2,3) - mat(2,2)*mat(1,3);
       cofactor(3,2)= -( mat(1,1)*mat(2,3) - mat(2,1)*mat(1,3) );
       cofactor(3,3)=    mat(1,1)*mat(2,2) - mat(2,1)*mat(1,2);

       det= mat(1,1)*cofactor(1,1) +  mat(1,2)*cofactor(1,2) + mat(1,3)*cofactor(1,3);

       if( fabs(det) < EPSILON)
       {
         return 0;
       }

       result(1,1)= cofactor(1,1)/det;
       result(1,2)= cofactor(2,1)/det;
       result(1,3)= cofactor(3,1)/det;

       result(2,1)= cofactor(1,2)/det;
       result(2,2)= cofactor(2,2)/det;
       result(2,3)= cofactor(3,2)/det;

       result(3,1)= cofactor(1,3)/det;
       result(3,2)= cofactor(2,3)/det;
       result(3,3)= cofactor(3,3)/det;
    break;
    //the inversion algorithm for arbitrary matrix size (Gauss-Jordan elimination with full pivoting)
    default: 
       for(i=0; i<dim; ++i)
       {  
         a[i][0]=0; 
         a[0][i]=0;
       }
       for(i=1; i<=dim; ++i)
       {  
         for(j=1; j<=dim; j++)
         {  
           //a[i][j]= M.data[j-1 + dim*(i-1)];
           a[i][j]= mat(i,j);
         }
       }
    
       for(i=1; i<=dim; ++i)
       {
         for(j=1; j<=dim; j++)
         {
           t[i][j]= a[i][j];
           if(i==j)
              t[i][j+dim]= 1.0;
           else
              t[i][j+dim]= 0.0;
         }
      }

      err= 1;
      k= 1;
      while(err==1 && k<=dim)
      {
        max= fabs(t[k][k]);
        l= k;
        for(i=k+1; i<=dim; ++i)
        {
          if(max<fabs(t[i][k]))
          {
            max= fabs(t[i][k]);
            l= i;
          }
        }
        if(max!=0)
        {
           for(j=k; j<=2*dim; j++)
           {
             max= t[k][j];
             t[k][j]= t[l][j];
             t[l][j]= max;
           }
           pivot= t[k][k];
           for(j=k+1; j<=2*dim; j++)
           {  t[k][j]/= pivot; }
           for(i=1; i<=dim; ++i)
           { 
              if(i!=k)
              {
                coef= t[i][k];
                for(j=k+1; j<=2*dim; j++)
                {  t[i][j]-= coef*t[k][j]; }
              }
           }
        }
        else
        { err= 0; }
        k++;
      }
      for(i=1; i<=dim; ++i)
      { 
         for(j=1; j<=dim; j++)
         {  b[i][j]= t[i][j+dim]; }
      }

      if(err==0)
      {
        result.setToZero();
        printf("%s: %d: inverse(Matrix &, Matrix &): non invertible matrix.\n",__FILE__,__LINE__);
        return 0;
      }


      for(i=1; i<=dim; ++i)
      {
        for(j=1; j<=dim; j++)
	{ 
          result(i,j)= b[i][j];
	}
      }

    break;
  }


  return 1;
}

int Matrix::setToRandom(float min, float max)
{
  unsigned int i, j;

  for(i=1; i<=nbRows_; ++i)
  {
    for(j=1; j<=nbColumns_; j++)
    {
      (*this)(i,j)=  randomNumber(min, max);
    }
  }

  return 1;
}




//! Computes the distance between a point P and the point C of the triangle (p0p1p2)
//! such as the distance from P to C is minimal.
//! \param p  the point
//! \param p0 triangle's first point
//! \param p1 triangle's second point
//! \param p2 triangle's third point
//! \param closestPoint the closest point to p on the triangle
//! \return the minimal distance between the point and the triangle
//! This function is adapted from the function "DistVector3Triangle3"
//! from "Wild Magic Library" (WM3), http://www.geometrictools.com by David Eberly.
float pointToTriangleDistance(const Vector3D &p, const Vector3D &p0, const Vector3D &p1, const Vector3D &p2, Vector3D &closestPoint)
{
    float fA00, fA01, fA11, fB0, fB1, fC;
    float fDet, fS, fT, fSqrDistance;
    Vector3D diff, edge0, edge1;

    diff= p0 - p;
    edge0= p1 - p0;
    edge1= p2 - p0;

    fA00= dotProduct(edge0, edge0);
    fA01= dotProduct(edge0, edge1);
    fA11= dotProduct(edge1, edge1);
    fB0 = dotProduct(diff, edge0);
    fB1 = dotProduct(diff, edge1);
    fC  = dotProduct(diff, diff);

    fDet = fabs(fA00*fA11-fA01*fA01);
    fS = fA01*fB1-fA11*fB0;
    fT = fA01*fB0-fA00*fB1;


    if (fS + fT <= fDet)
    {
        if (fS < 0.0)
        {
            if (fT < 0.0)  // region 4
            {
                if (fB0 < 0.0)
                {
                    fT = 0.0;
                    if (-fB0 >= fA00)
                    {
                        fS = 1.0;
                        fSqrDistance = fA00+(2.0)*fB0+fC;
                    }
                    else
                    {
                        fS = -fB0/fA00;
                        fSqrDistance = fB0*fS+fC;
                    }
                }
                else
                {
                    fS = 0.0;
                    if (fB1 >= 0.0)
                    {
                        fT = 0.0;
                        fSqrDistance = fC;
                    }
                    else if (-fB1 >= fA11)
                    {
                        fT = 1.0;
                        fSqrDistance = fA11+(2.0)*fB1+fC;
                    }
                    else
                    {
                        fT = -fB1/fA11;
                        fSqrDistance = fB1*fT+fC;
                    }
                }
            }
            else  // region 3
            {
                fS = 0.0;
                if (fB1 >= 0.0)
                {
                    fT = 0.0;
                    fSqrDistance = fC;
                }
                else if (-fB1 >= fA11)
                {
                    fT = 1.0;
                    fSqrDistance = fA11+(2.0)*fB1+fC;
                }
                else
                {
                    fT = -fB1/fA11;
                    fSqrDistance = fB1*fT+fC;
                }
            }
        }
        else if (fT < 0.0)  // region 5
        {
            fT = 0.0;
            if (fB0 >= 0.0)
            {
                fS = 0.0;
                fSqrDistance = fC;
            }
            else if (-fB0 >= fA00)
            {
                fS = 1.0;
                fSqrDistance = fA00+(2.0)*fB0+fC;
            }
            else
            {
                fS = -fB0/fA00;
                fSqrDistance = fB0*fS+fC;
            }
        }
        else  // region 0
        {
            // minimum at interior point
            double fInvDet = (1.0)/fDet;
            fS *= fInvDet;
            fT *= fInvDet;
            fSqrDistance = fS*(fA00*fS+fA01*fT+(2.0)*fB0) +
                fT*(fA01*fS+fA11*fT+(2.0)*fB1)+fC;
        }
    }
    else
    {
        double fTmp0, fTmp1, fNumer, fDenom;

        if (fS < 0.0)  // region 2
        {
            fTmp0 = fA01 + fB0;
            fTmp1 = fA11 + fB1;
            if (fTmp1 > fTmp0)
            {
                fNumer = fTmp1 - fTmp0;
                fDenom = fA00-2.0f*fA01+fA11;
                if (fNumer >= fDenom)
                {
                    fS = 1.0;
                    fT = 0.0;
                    fSqrDistance = fA00+(2.0)*fB0+fC;
                }
                else
                {
                    fS = fNumer/fDenom;
                    fT = 1.0 - fS;
                    fSqrDistance = fS*(fA00*fS+fA01*fT+2.0f*fB0) +
                        fT*(fA01*fS+fA11*fT+(2.0)*fB1)+fC;
                }
            }
            else
            {
                fS = 0.0;
                if (fTmp1 <= 0.0)
                {
                    fT = 1.0;
                    fSqrDistance = fA11+(2.0)*fB1+fC;
                }
                else if (fB1 >= 0.0)
                {
                    fT = 0.0;
                    fSqrDistance = fC;
                }
                else
                {
                    fT = -fB1/fA11;
                    fSqrDistance = fB1*fT+fC;
                }
            }
        }
        else if (fT < 0.0)  // region 6
        {
            fTmp0 = fA01 + fB1;
            fTmp1 = fA00 + fB0;
            if (fTmp1 > fTmp0)
            {
                fNumer = fTmp1 - fTmp0;
                fDenom = fA00-(2.0)*fA01+fA11;
                if (fNumer >= fDenom)
                {
                    fT = 1.0;
                    fS = 0.0;
                    fSqrDistance = fA11+(2.0)*fB1+fC;
                }
                else
                {
                    fT = fNumer/fDenom;
                    fS = 1.0 - fT;
                    fSqrDistance = fS*(fA00*fS+fA01*fT+(2.0)*fB0) +
                        fT*(fA01*fS+fA11*fT+(2.0)*fB1)+fC;
                }
            }
            else
            {
                fT = 0.0;
                if (fTmp1 <= 0.0)
                {
                    fS = 1.0;
                    fSqrDistance = fA00+(2.0)*fB0+fC;
                }
                else if (fB0 >= 0.0)
                {
                    fS = 0.0;
                    fSqrDistance = fC;
                }
                else
                {
                    fS = -fB0/fA00;
                    fSqrDistance = fB0*fS+fC;
                }
            }
        }
        else  // region 1
        {
            fNumer = fA11 + fB1 - fA01 - fB0;
            if (fNumer <= 0.0)
            {
                fS = 0.0;
                fT = 1.0;
                fSqrDistance = fA11+(2.0)*fB1+fC;
            }
            else
            {
                fDenom = fA00-2.0f*fA01+fA11;
                if (fNumer >= fDenom)
                {
                    fS = 1.0;
                    fT = 0.0;
                    fSqrDistance = fA00+(2.0)*fB0+fC;
                }
                else
                {
                    fS = fNumer/fDenom;
                    fT = 1.0 - fS;
                    fSqrDistance = fS*(fA00*fS+fA01*fT+(2.0)*fB0) +
                        fT*(fA01*fS+fA11*fT+(2.0)*fB1)+fC;
                }
            }
        }
    }


    // account for numerical round-off error
    if (fSqrDistance < 0.0)
    {
        fSqrDistance = 0.0;
    }


    int i;
    for(i= 0; i<3; ++i)
    { closestPoint[i]= p0[i] + fS*edge0[i] + fT*edge1[i]; }


    return sqrt(fSqrDistance);
}

