
#include <list>
#include <map>
#include <algorithm>
#include <fstream>
#if defined(ENABLE_DISPLAY)
#include <GL/gl.h>
#endif
#include <vector3D.h>
#include <body.h>
#include <file_reader.h>
#include <coldman.h>
#include <graphics.h>
#include <workspace.h>

Vector3D mgd(float q1, float q2, float q3)
{
  float l0, l1, l2, l3;
  Vector3D p;

  l0  =   0.0855; 
  l1  =   0.067816; 
  l2  =   0.029980; 
  l3  =   0.029;

  p.x= -sin(q1)*(  l1*cos(q2) + l2*cos(q2+q3) + l3*cos(q2+2*q3)  ); 
  p.y=  cos(q1)*(  l1*cos(q2) + l2*cos(q2+q3) + l3*cos(q2+2*q3)  );
  p.z= -l1*sin(q2) - l2*sin(q2+q3) - l3*sin(q2+2*q3);

  return p;
}

//! Computes the position of the fingertip center point of the Schunk Anthropomorphic Hand finger.
//! \param q1 first joint parameter (controls the (palm)-(proximal phalanx) joint first DOF (abduction))
//! \param q2 second joint (controls the (palm)-(proximal phalanx) joint second DOF (subduction))
//! \param q3 third joint parameter (controls the DOFs of the last two joints that are coupled)
//! \param length1 lenght of the first phalanx
//! \param length2 lenght of the second phalanx
//! \param length3 lenght of the third phalanx
//! \return the position of the fingertip center point
Vector3D SAHfingerForwardKinematics(float q1, float q2, float q3, float length1, float length2, float length3)
{
  Vector3D p;
/*
  l0  =   0.0855; 
  l1  =   0.067816; 
  l2  =   0.029980; 
  l3  =   0.029;*/

  p.x= -sin(q1)*(  length1*cos(q2) + length2*cos(q2+q3) + length3*cos(q2+2*q3)  ); 
  p.y=  cos(q1)*(  length1*cos(q2) + length2*cos(q2+q3) + length3*cos(q2+2*q3)  );
  p.z= -length1*sin(q2) - length2*sin(q2+q3) - length3*sin(q2+2*q3);

  return p;
}

/*
int drawSampledWorkspace(HTMatrix Tbase)
{
  float l0, l1, l2, l3;
  float q1min, q1max, q2min, q2max, q3min, q3max, dq;
  float q1, q2, q3;
  unsigned int i, j, k, n1, n2, n3;
  Vector3D p1, p2, p3;
  GLfloat mat[16];

  l0  =   0.0855; 
  l1  =   0.067816; 
  l2  =   0.029980; 
  l3  =   0.029;

  q1min= -20*DEGTORAD;
  q1max=  20*DEGTORAD;
  q2min= -19*DEGTORAD;
  q2max=  90*DEGTORAD;
  q3min=   0*DEGTORAD;
  q3max=  90*DEGTORAD;

  dq= 5*DEGTORAD;

  n1= (unsigned int) ( (q1max-q1min)/dq );
  n2= (unsigned int) ( (q2max-q2min)/dq );
  n3= (unsigned int) ( (q3max-q3min)/dq );

  Tbase.convertToOpenGLFormat(mat);
  glPointSize(3);


  glPushMatrix();
    glMultMatrixf(mat);
    glRotatef(90, 0, 1, 0);
    glBegin(GL_POINTS);
      for(i=0; i<n1; ++i)
      {
        q1= q1min + i*dq;
    
        p1.x= -sin(q1)*( l1*cos(q2) ); 
        p1.y=  cos(q1)*( l1*cos(q2) );
        p1.z= -l1*sin(q2);
    
        for(j=0; j<n2; j++)
        {
          q2= q2min + j*dq;
    
    
    
        p2.x= -sin(q1)*( l1*cos(q2) + l2*cos(q2+q3) ); 
        p2.y=  cos(q1)*( l1*cos(q2) + l2*cos(q2+q3) );
        p2.z= -l1*sin(q2) - l2*sin(q2+q3);
    
    
          for(k=0; k<n3; k++)
          {
            q3= q3min + k*dq;
    
            p3= mgd(q1, q2, q3);
            glVertex3f(p3.x, p3.y, p3.z);
          }
        }
      }
    glEnd();

glColor3f(1, 0, 0);
    glBegin(GL_LINE_LOOP);
      q1= 0;
      q3= q3min;
      for(i=0; i<n2; ++i)
      {
        q2= q2min + i*dq;
        p3= mgd(q1, q2, q3);
        glVertex3f(p3.x, p3.y, p3.z);
      }

      q1= 0;
      q2= q2max;
      for(i=0; i<n3; ++i)
      {
        q3= q3min + i*dq;
        p3= mgd(q1, q2, q3);
        glVertex3f(p3.x, p3.y, p3.z);
      }

      q1= 0;
      q3= q3max;
      for(i=0; i<n2; ++i)
      {
        q2= q2max - i*dq;
        p3= mgd(q1, q2, q3);
        glVertex3f(p3.x, p3.y, p3.z);
      }

      q1= 0;
      q2= q2min;
      for(i=0; i<n3; ++i)
      {
        q3= q3max - i*dq;
        p3= mgd(q1, q2, q3);
        glVertex3f(p3.x, p3.y, p3.z);         
      }
    glEnd();

  glPopMatrix();

 return 1;
}


Vector3D computeContour(float q1, unsigned int n2, unsigned int n3, unsigned int i)
{
  float q2min, q2max, q3min, q3max, dq2, dq3;
  float q2, q3;


  q2min= -19*DEGTORAD;
  q2max=  90*DEGTORAD;
  q3min=   0*DEGTORAD;
  q3max=  90*DEGTORAD;

  dq2= (q2max - q2min)/( (float) n2 );
  dq3= (q3max - q3min)/( (float) n3 );

  if(i < n2)
  {
    q2= q2min + i*dq2;
    q3= q3min;   
    return mgd(q1, q2, q3);
  }

  if(i < n2 + n3)
  {
    q2= q2max;
    q3= q3min + (i-n2)*dq3;   
    return mgd(q1, q2, q3);
  }  


  if(i < 2*n2 + n3)
  {
    q2= q2max - (i-n2-n3)*dq2;
    q3= q3max;   
    return mgd(q1, q2, q3);
  }  

  q2= q2min;
  q3= q3max - (i-2*n2-n3)*dq3;  
  return mgd(q1, q2, q3);
}

int drawContour(HTMatrix Tbase)
{
  float q1min, q1max, q2min, q2max, q3min, q3max, dq;
  unsigned int i, n1, n2, n3, n;
  Vector3D p1, p2, p3, p4, e1, e2, pn;
  GLfloat mat[16];

  q1min= -20*DEGTORAD;
  q1max=  20*DEGTORAD;
  q2min= -19*DEGTORAD;
  q2max=  90*DEGTORAD;
  q3min=   0*DEGTORAD;
  q3max=  90*DEGTORAD;

  dq= 5*DEGTORAD;

  n1= (unsigned int) ( (q1max-q1min)/dq );
  n2= (unsigned int) ( (q2max-q2min)/dq );
  n3= (unsigned int) ( (q3max-q3min)/dq );

  n= 2*(n2+n3);
//Vector3D computeContour(float q1, unsigned int n2, unsigned int n3, unsigned int i)
  Tbase.convertToOpenGLFormat(mat);

  glPushMatrix();
    glMultMatrixf(mat);
    glRotatef(90, 0, 1, 0);

    glColor3f(0, 0.4, 1); 
    glBegin(GL_TRIANGLE_STRIP);
    for(i=0; i<n; ++i)
    {
      p1= computeContour(q1min, n2, n3, i);
      p2= computeContour(q1min, n2, n3, n-1-i);
      glVertex3f(p1.x, p1.y, p1.z);
      glVertex3f(p2.x, p2.y, p2.z);
    }
    glEnd();
 glColor3f(0, 1.0, 0.4); 
    glBegin(GL_TRIANGLE_STRIP);
    for(i=0; i<n; ++i)
    {
      p1= computeContour(q1max, n2, n3, i);
      p2= computeContour(q1max, n2, n3, n-1-i);
      glVertex3f(p1.x, p1.y, p1.z);
      glVertex3f(p2.x, p2.y, p2.z);
    }
    glEnd();
  glPopMatrix();

  return 1;
}


int drawSampledWorkspace2(HTMatrix Tbase)
{
  float q1min, q1max, q2min, q2max, q3min, q3max, dq;
  unsigned int i, i1, i2, j, j1, j2, n1, n2, n3, n;
  Vector3D p1, p2, p3, p4, e1, e2, pn;
  GLfloat mat[16];
  TriangleMesh mesh;


  q1min= -20*DEGTORAD;
  q1max=  20*DEGTORAD;
  q2min= -19*DEGTORAD;
  q2max=  90*DEGTORAD;
  q3min=   0*DEGTORAD;
  q3max=  90*DEGTORAD;

  dq= 5*DEGTORAD;

  n1= (unsigned int) ( (q1max-q1min)/dq );
  n2= (unsigned int) ( (q2max-q2min)/dq );
  n3= (unsigned int) ( (q3max-q3min)/dq );

  Tbase.convertToOpenGLFormat(mat);
  glPointSize(3);

  n= 2*(n2+n3);

  glPushMatrix();
    glMultMatrixf(mat);
    glRotatef(90, 0, 1, 0);

    glColor3f(0, 0.4, 1);
    glBegin(GL_TRIANGLES);
    for(i=0; i<n1-1; ++i)
    {
      i1= i;
      i2= i+1;
     // if(i2==n1)
      //{ i2= 0; }

      for(j=0; j<2*(n2+n3); j++)
      {
        j1= j;
        j2= j+1;
        if(j2==2*(n2+n3))
        {  j2= 0; }

       // q1= q1min + 
        p1= computeContour(q1min + i1*dq, n2, n3, j1);
        p2= computeContour(q1min + i1*dq, n2, n3, j2);
        p3= computeContour(q1min + i2*dq, n2, n3, j2);
        p4= computeContour(q1min + i2*dq, n2, n3, j1);
  
        mesh.addVertex(p1.x, p1.y, p1.z);
        mesh.addVertex(p2.x, p2.y, p2.z);

//         e1= p2 - p1;
//         e1.normalize();
//         e2= p4 - p1;
//         e2.normalize();
//         pn= crossProduct(e1, e2);
//         pn.normalize();
      //  glNormal3f(pn.x, pn.y, pn.z);
        glVertex3f(p1.x, p1.y, p1.z);
        glVertex3f(p2.x, p2.y, p2.z);
        glVertex3f(p4.x, p4.y, p4.z);


//         e1= p3 - p2;
//         e1.normalize();
//         e2= p4 - p2;
//         e2.normalize();
//         pn= crossProduct(e1, e2);
//         pn.normalize();
        glVertex3f(p2.x, p2.y, p2.z);
        glVertex3f(p3.x, p3.y, p3.z);
        glVertex3f(p4.x, p4.y, p4.z);
      }
    }
    glEnd();

  glPopMatrix();

 return 1;
}
*/

//! Computes the workspace of a SAHand finger.
//! \param dq discretization step for the joint parameters
//! \param q1min minimal bound on the first joint
//! \param q1max maximal bound on the first joint
//! \param q2min minimal bound on the second joint
//! \param q2max maximal bound on the second joint
//! \param q3min minimal bound on the third joint (that is coupled with the fourth one)
//! \param q3max maximal bound on the third joint (that is coupled with the fourth one)
//! \param length1 lenght of the first phalanx
//! \param length2 lenght of the second phalanx
//! \param length3 lenght of the third phalanx
//! \return 1 in case of success, 0 otherwise
int SAHfingerWorkspace(float dq, float q1min, float q1max, float q2min, float q2max, float q3min, float q3max, float length1, float length2, float length3)
{
  std::list<Vector2D>::iterator previous, current, next;
  std::list<Vector2D> contour, contourPlus, contourMinus;
  unsigned int i, n1, n2, n3;
  float q1, q2, q3;
  Vector2D p2D, axis, firstMinus, lastMinus, firstPlus, lastPlus;
  Vector3D p;
  TriangleMesh mesh;

  n1= (unsigned int) floor( (q1max-q1min)/dq );
  n2= (unsigned int) floor( (q2max-q2min)/dq );
  n3= (unsigned int) floor( (q3max-q3min)/dq );

  //compute the curve generated by the two last joints, that will be swept to give the workspace volume:
  q1= q1min;
  q3= q3min;
  for(i=0; i<n2; ++i)
  {
    q2= q2min + i*dq;
    p= SAHfingerForwardKinematics(q1, q2, q3, length1, length2, length3);
    p2D.x= p.y;
    p2D.y= p.z;
    contour.push_back(p2D);
  }  

  q2= q2max;
  for(i=0; i<n3; ++i)
  {
    q3= q3min + i*dq;
    p= SAHfingerForwardKinematics(q1, q2, q3, length1, length2, length3);
    p2D.x= p.y;
    p2D.y= p.z;
    contour.push_back(p2D);
  }  

  q3= q3max;
  for(i=0; i<n2; ++i)
  {
    q2= q2max - i*dq;
    p= SAHfingerForwardKinematics(q1, q2, q3, length1, length2, length3);
    p2D.x= p.y;
    p2D.y= p.z;
    contour.push_back(p2D);
  }  

  q2= q2min;
  for(i=0; i<n3; ++i)
  {
    q3= q3max - i*dq;
    p= SAHfingerForwardKinematics(q1, q2, q3, length1, length2, length3);
    p2D.x= p.y;
    p2D.y= p.z;
    contour.push_back(p2D);
  }  

  //remove consecutive vertices that are too close (possible duplicate vertices):
  current= contour.begin();
  while(current!=contour.end())
  {
     if(current==contour.begin())
     {  previous= contour.end();}
     else 
     {  previous= current;  }
     previous--;

     next= current;
     next++;
     if(next==contour.end())
     {  next= contour.begin();  }

     if( norm(*current - *previous) < EPSILON )
     {
       current= contour.erase(current);
       continue;
     }

     current++;
  }


  //cut the 2D curve in two: one part on the right of the rotation axis, the other one on the left:
  axis.set(0, 1);
  p2D= axis.orthogonalVector();
  std::list<Vector2D>::iterator startPlus, endPlus, startMinus, endMinus;
  for(current=contour.begin(); current!=contour.end(); current++)
  {
     if(current==contour.begin())
     {  previous= contour.end();}
     else 
     {  previous= current;  }
     previous--;

     next= current;
     next++;
     if(next==contour.end())
     {  next= contour.begin();  }


     if( (dotProduct(*previous, p2D) < 0.0) && (dotProduct(*current, p2D) > 0.0)) 
     {
       startPlus= current;
       endMinus= current;  
     }
     if( (dotProduct(*previous, p2D) > 0.0) && (dotProduct(*current, p2D) < 0.0)) 
     {
       startMinus= current; 
       endPlus= current;
     }
  }


  current= startMinus;
  while(current!=endMinus)
  {
    contourMinus.push_back(*current);
    current++;
    if(current==contour.end())
    {
      current= contour.begin();
    }
  }

  current= startPlus;
  while(current!=endPlus)
  {
    contourPlus.push_back(*current);
    current++;
    if(current==contour.end())
    {
      current= contour.begin();
    }
  }



  // add points where the curve was cut (on the roation axis): 
  firstMinus= contourMinus.front();
  lastMinus = contourMinus.back();
  firstPlus = contourPlus.front();
  lastPlus =  contourPlus.back();

  if( fabs(contourMinus.front().x) > EPSILON)
  { 
    if( fabs(lastPlus.x) <  fabs(firstMinus.x) )
    {  p2D.set( 0.0, lastPlus.y);   } 
    else
    {  p2D.set( 0.0, firstMinus.y); } 
    contourMinus.push_front(p2D); 
  }

  if( fabs(contourMinus.back().x) > EPSILON)
  { 
    if( fabs(firstPlus.x) <  fabs(lastMinus.x) )
    {  p2D.set( 0.0, firstPlus.y);   } 
    else
    {  p2D.set( 0.0, lastMinus.y); } 
    contourMinus.push_back(p2D); 
  }


  if( fabs(contourPlus.front().x) > EPSILON)
  { 
    if( fabs(lastMinus.x) <  fabs(firstPlus.x) )
    {  p2D.set( 0.0, lastMinus.y);   } 
    else
    {  p2D.set( 0.0, firstPlus.y); } 
    contourPlus.push_front(p2D); 
  }

  if( fabs(contourPlus.back().x) > EPSILON)
  { 
    if( fabs(firstMinus.x) <  fabs(lastPlus.x) )
    {  p2D.set( 0.0, firstMinus.y);   } 
    else
    {  p2D.set( 0.0, lastPlus.y); } 
    contourPlus.push_back(p2D); 
  }


  printf("p= [\n");
  for(current=contour.begin(); current!=contour.end(); current++)
  {  printf("%f %f ;\n", current->x, current->y);  }
  printf("];\n");

  printf("pplus= [\n");
  for(current=contourPlus.begin(); current!=contourPlus.end(); current++)
  {  printf("%f %f ;\n", current->x, current->y);  }
  printf("];\n");


  printf("pminus= [\n");
  for(current=contourMinus.begin(); current!=contourMinus.end(); current++)
  {  printf("%f %f ;\n", current->x, current->y);  }
  printf("];\n");

  //float alpha0= -10*DEGTORAD;
  //float alpha1= 40*DEGTORAD;

  // build the revolution surfaces:
  TriangleMesh surf;
  surfaceOfRevolution(contourPlus, q1min, q1max, 4*DEGTORAD, 0.001, surf);
  surf.saveAsOBJfile("surf.obj", 1.0);   

  TriangleMesh surf2;
  //surfaceOfRevolution(contourMinus, M_PI-alpha1, M_PI-alpha0, 4*DEGTORAD, 0.001, surf2);
  surfaceOfRevolution(contourMinus, q1min, q1max, 4*DEGTORAD, 0.001, surf2);

  for(i=0; i<surf2.nbVertices(); ++i)
  {
    //surf2.vertices[i].z= -surf2.vertices[i].z;
  }
  //surf2.invertTriangleNormals();
  surf2.saveAsOBJfile("surf2.obj", 1.0);

 

  return 1;
}





void printM(float M[3][3], const char *name)
{

 if(name[0] != '\0') 
    printf("matrix %s\n", name);
  
 printf("%f  %f  %f  \n", M[0][0], M[0][1], M[0][2]);
 printf("%f  %f  %f  \n", M[1][0], M[1][1], M[1][2]);
 printf("%f  %f  %f  \n", M[2][0], M[2][1], M[2][2]);
}


//! From a sample set of two variables (x,y), computes the linear dependency between the variable y and the variable y.
//! The searched relationship has the form: y= a*x + b
//! \param points a sample set of two variables (represented by 2D vectors (x,y))
//! \param result a 2D vector that will be filled with the computed parameters of the linear combination (a, b)  
//! such that y= a*x + b)
//! \return 1 in case of success, 0 otherwise (y is independent from x)
//! Principle:
//! y= a*x + b
//!        (  x1   1  )           (  y1  )
//!        (  x2   1  )           (  y2  )
//!  X =   (  |    |  )     Y =   (  |   )
//!        (  xn   1  )           (  yn  )
//!
//!  A =   (  a  )
//!        (  b  )
//!
//!  Y = X*A
//!  -> A = inv(X'*X)*X'*Y
int linearRegression2D(std::vector<Vector2D> &points, Vector2D &result)
{
  unsigned int i;
  float T[2], M[2][2], invM[2][2], det;

  if(points.size() < 2)
  {
    printf("%s: %d: linearRegression2D(): at least 2 points are needed.\n",__FILE__,__LINE__);
    return 0;
  }

  // M= X'*X;
  M[0][0]= M[0][1]= 0.0;
  M[1][0]= M[1][1]= 0.0;

  for(i=0; i<points.size(); ++i)
  {
    M[0][0]+= points[i].x*points[i].x;
    M[0][1]+= points[i].x;
  }

  M[1][0]= M[0][1];
  M[1][1]= points.size();

  det= M[0][0]*M[1][1] - M[0][1]*M[1][0];

  if( fabs(det) < EPSILON )
  {
    return 0;
  }

  invM[0][0]=  M[1][1]/det;
  invM[0][1]= -M[0][1]/det;
  invM[1][0]= -M[1][0]/det;
  invM[1][1]=  M[0][0]/det;


  // T= X'*Y;
  T[0]= T[1]= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    T[0]+= points[i].x*points[i].y;
    T[1]+= points[i].y;
  }

  result.x= invM[0][0]*T[0] + invM[0][1]*T[1];
  result.y= invM[1][0]*T[0] + invM[1][1]*T[1];


  if( isnan(result.x) || isnan(result.y) || isinf(result.x) || isinf(result.y) )
  {
    printf("M= [ %f %f  ;  %f %f ] \n", M[0][0], M[0][1], M[1][0], M[1][1]);
    printf("invM= [ %f %f  ;  %f %f ] \n", invM[0][0], invM[0][1], invM[1][0], invM[1][1]);
    printf("%s: %d: linearRegression2D(): computation error.\n",__FILE__,__LINE__);
    return 0;
  }

  return 1;
}


//! From a sample set of three variables (x,y,z), computes the linear dependency between the variable z and the other two variables (x and y).
//! The searched relationship has the form: z= a*x + b*y + c.
//! \param points a sample set of three variables (represented by 3D vectors (x,y,z))
//! \param result a 3D vector that will be filled with the computed parameters of the linear combination (a, b, c) 
//! such that z= a*x + b*y + c)
//! \return 1 in case of success, 0 otherwise (z is independent from x and y)
//! Principle:
//! z= a*x + b*y + c
//!        (  x1   y1   1  )           (  z1  )
//!        (  x2   y2   1  )           (  z2  )
//!  X =   (  |    |    |  )     Y =   (  |   )
//!        (  xn   yn   1  )           (  zn  )
//!
//!        (  a  )
//!  A =   (  b  )
//!        (  c  )
//!
//!  Y = X*A
//!  -> A = inv(X'*X)*X'*Y
int linearRegression3D(std::vector<Vector3D> &points, Vector3D &result)
{
  unsigned int i;
  float T[3], M[3][3], invM[3][3], cofactor[3][3], det;

  if(points.size() < 3)
  {
    printf("%s: %d: linearRegression3D(): at least 3 points are needed.\n",__FILE__,__LINE__);
    return 0;
  }

  // M= X'*X;
  M[0][0]= M[0][1]=  M[0][2]= 0.0;
  M[1][0]= M[1][1]=  M[1][2]= 0.0;
  M[2][0]= M[2][1]=  M[2][2]= 0.0;
  
  for(i=0; i<points.size(); ++i)
  {
    M[0][0]+= points[i].x*points[i].x;
    M[0][1]+= points[i].x*points[i].y;
    M[0][2]+= points[i].x;   

    M[1][1]+= points[i].y*points[i].y;
    M[1][2]+= points[i].y; 
  }
 
  M[1][0]= M[0][1];
  M[2][0]= M[0][2];
  M[2][1]= M[1][2];
  M[2][2]= points.size();

  cofactor[0][0]=  ( M[1][1]*M[2][2] - M[2][1]*M[1][2] );
  cofactor[0][1]= -( M[1][0]*M[2][2] - M[2][0]*M[1][2] );
  cofactor[0][2]=  ( M[1][0]*M[2][1] - M[2][0]*M[1][1] );

  cofactor[1][0]= -( M[0][1]*M[2][2] - M[2][1]*M[0][2] );
  cofactor[1][1]=  ( M[0][0]*M[2][2] - M[2][0]*M[0][2] );
  cofactor[1][2]= -( M[0][0]*M[2][1] - M[2][0]*M[0][1] );

  cofactor[2][0]=  ( M[0][1]*M[1][2] - M[1][1]*M[0][2] );
  cofactor[2][1]= -( M[0][0]*M[1][2] - M[1][0]*M[0][2] );
  cofactor[2][2]=  ( M[0][0]*M[1][1] - M[1][0]*M[0][1] );

  det= M[0][0]*cofactor[0][0] + M[0][1]*cofactor[0][1] + M[0][2]*cofactor[0][2];

  if( fabs(det) < EPSILON )
  {
    return 0;
  }

  invM[0][0]= cofactor[0][0]/det;
  invM[0][1]= cofactor[1][0]/det;
  invM[0][2]= cofactor[2][0]/det;
  invM[1][0]= cofactor[0][1]/det;
  invM[1][1]= cofactor[1][1]/det;
  invM[1][2]= cofactor[2][1]/det;
  invM[2][0]= cofactor[0][2]/det;
  invM[2][1]= cofactor[1][2]/det;
  invM[2][2]= cofactor[2][2]/det;


  // T= X'*Y;
  T[0]= T[1]= T[2]= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    T[0]+= points[i].x*points[i].z;
    T[1]+= points[i].y*points[i].z;
    T[2]+= points[i].z;
  }

  result.x= invM[0][0]*T[0] + invM[0][1]*T[1] + invM[0][2]*T[2];
  result.y= invM[1][0]*T[0] + invM[1][1]*T[1] + invM[1][2]*T[2];
  result.z= invM[2][0]*T[0] + invM[2][1]*T[1] + invM[2][2]*T[2];
  
  if( isnan(result.x) || isnan(result.y) || isnan(result.z) || isinf(result.x) || isinf(result.y) || isinf(result.z) )
  {
    printf("%s: %d: linearRegression3D(): computation error.\n",__FILE__,__LINE__);
    return 0;
  }

  return 1;
}

//! Computes the best fitting plane (in least-square meaning) for a set of 3D points.
//! Four possibilities are tested and the one with the smallest error is chosen.
//! These possibilities are:
//!  - z= f(x,y)
//!  - z does no depend on x and y, y= f(y) -> plane with normal parallel to 0xy plane
//!  - plane x= constant
//!  - plane y= constant
//! NB: if the points are on a line, the function will return the equation of an arbitrary plane containing this line.
//! \param points the input points
//! \param result the computed plane equation parameters ((a,b,c,d) such that a*x + b*y + c*z + d = 0)
//! \return 1 in case of success, 0 otherwise
int bestFittingPlane(std::vector<Vector3D> &points, float result[4])
{
  unsigned int i;
  float meanX, meanY, varX, varY;
  bool solution1, solution2;
  float error1, error2, error3;
  float param1[4]= {0,0,0,0}, param2[4]= {0,0,0,0}, param3[4]= {0,0,0,0};
  Vector2D param2D;
  Vector3D param3D;
  std::vector<Vector2D> points2D;
  std::vector<Vector2D> points2;


  // z= f(x,y):
  if( linearRegression3D(points, param3D)==1 )
  {
    solution1= true;
    param1[0]= param3D.x;
    param1[1]= param3D.y;
    param1[2]= -1.0;
    param1[3]= param3D.z; 
  }
  else
  {
    solution1= false;    
  }


  // y= f(x)
  points2D.resize(points.size());
  for(i=0; i<points.size(); ++i)
  {
    points2D[i].set(points[i].x, points[i].y);
  }
  if( linearRegression2D(points2D, param2D)==1 )
  {
    solution2= true;
    param2[0]= param2D.x;
    param2[1]= -1.0;
    param2[2]= 0.0;
    param2[3]= param2D.y;    
  }
  else
  {
    solution2= false;    
  }


  // x and y are independent:
  // plane x= constant or y= constant
  meanX= 0.0;
  meanY= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    meanX+= points[i].x;
    meanY+= points[i].y;
  }
  meanX/= (float) (points.size());
  meanY/= (float) (points.size());

  varX= 0.0;
  varY= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    varX+= (points[i].x - meanX)*(points[i].x - meanX);
    varY+= (points[i].y - meanY)*(points[i].y - meanY);
  }
  varX= sqrt(varX);
  varY= sqrt(varY);
  varX/= (float) (points.size());
  varY/= (float) (points.size());

  //choose the plane with smallest variance:
  if(varX > varY)
  {
    param3[0]= 0.0;
    param3[1]= 1.0;
    param3[2]= 0.0;
    param3[3]= -meanY;
  }
  else
  {
    param3[0]= 1.0;
    param3[1]= 0.0;
    param3[2]= 0.0;
    param3[3]= -meanX;
  }


  error1= error2= error3= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    if(solution1)
    {  error1+= pow(param1[0]*points[i].x + param1[1]*points[i].y + param1[2]*points[i].z + param1[3], 2);  }

    if(solution2)
    {  error2+= pow(param2[0]*points[i].x + param2[1]*points[i].y + param2[2]*points[i].z + param2[3], 2);  }

    error3+= pow(param3[0]*points[i].x + param3[1]*points[i].y + param3[2]*points[i].z + param3[3], 2);
  }
  error1/= (float) (points.size());
  error2/= (float) (points.size());
  error3/= (float) (points.size());

  if( solution1 )
  {
    if( (error1 < error2) && (error1 < error3) )
    {
      result[0]= param1[0];
      result[1]= param1[1];
      result[2]= param1[2];
      result[3]= param1[3];
      return 1;
    }
  }

  if( solution2 )
  {
    if( (error2 < error1) && (error2 < error3) )
    {
      result[0]= param2[0];
      result[1]= param2[1];
      result[2]= param2[2];
      result[3]= param2[3];
      return 1;
    }
  }

  result[0]= param3[0];
  result[1]= param3[1];
  result[2]= param3[2];
  result[3]= param3[3];

  return 1;

  if(error1 < error2)
  {
    if(error1 < error3)
    {
      result[0]= param1[0];
      result[1]= param1[1];
      result[2]= param1[2];
      result[3]= param1[3];
      return 1;
    }
    else
    {
      result[0]= param3[0];
      result[1]= param3[1];
      result[2]= param3[2];
      result[3]= param3[3];
      return 1;
    }
  }
  if(error2 < error3)
  {
    result[0]= param2[0];
    result[1]= param2[1];
    result[2]= param2[2];
    result[3]= param2[3];
    return 1;
  }
  else
  {
    result[0]= param3[0];
    result[1]= param3[1];
    result[2]= param3[2];
    result[3]= param3[3];
    return 1;
  }

  return 1;
}

#if defined(ENABLE_DISPLAY)
//! Draws (a part of) a plane from its equation.
//! \param a first parameter of the plane's equation (a*x + b*y + c*z + d = 0)
//! \param b second parameter of the plane's equation (a*x + b*y + c*z + d = 0)
//! \param c third parameter of the plane's equation (a*x + b*y + c*z + d = 0)
//! \param d fourth parameter of the plane's equation (a*x + b*y + c*z + d = 0)
//! \param size controls how large will be the displayed part
//! \return 1 in case of success, 0 otherwise
int drawPlane(float a, float b, float c, float d, float size)
{
  GLfloat p1[3], p2[3], p3[3], p4[3];

  if( fabs(a)<EPSILON && fabs(b)<EPSILON && fabs(c)<EPSILON )
  {
    printf("%s: %d: drawPlane(): plane is degenerate (a=%f b=%f c=%f).\n",__FILE__,__LINE__,a, b, c);    
    return 0;
  }

  if( fabs(b)<EPSILON && fabs(c)<EPSILON )
  { 
    p1[0]= -d/a;    p1[1]= -0.5*size;    p1[2]= -0.5*size;
    p2[0]= -d/a;    p2[1]=  0.5*size;    p2[2]= -0.5*size;
    p3[0]= -d/a;    p3[1]=  0.5*size;    p3[2]=  0.5*size;
    p4[0]= -d/a;    p4[1]= -0.5*size;    p4[2]=  0.5*size;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1;
  }  

  if( fabs(a)<EPSILON && fabs(c)<EPSILON )
  { 
    p1[0]= -0.5*size;    p1[1]= -d/b;    p1[2]= -0.5*size;
    p2[0]=  0.5*size;    p2[1]= -d/b;    p2[2]= -0.5*size;
    p3[0]=  0.5*size;    p3[1]= -d/b;    p3[2]=  0.5*size;
    p4[0]= -0.5*size;    p4[1]= -d/b;    p4[2]=  0.5*size;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1;
  }  

  if( fabs(a)<EPSILON && fabs(b)<EPSILON )
  { 
    p1[0]= -0.5*size;    p1[1]= -0.5*size;    p1[2]= -d/c;
    p2[0]=  0.5*size;    p2[1]= -0.5*size;    p2[2]= -d/c;
    p3[0]=  0.5*size;    p3[1]=  0.5*size;    p3[2]= -d/c;
    p4[0]= -0.5*size;    p4[1]=  0.5*size;    p4[2]= -d/c;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1;
  }  



  if( fabs(a)<EPSILON )
  { 
    p1[0]= -0.5*size;    p1[1]= -0.5*size;    p1[2]= -(b*p1[1] + d)/c;
    p2[0]= -0.5*size;    p2[1]=  0.5*size;    p2[2]= -(b*p2[1] + d)/c;
    p3[0]=  0.5*size;    p3[1]=  0.5*size;    p3[2]= -(b*p3[1] + d)/c;
    p4[0]=  0.5*size;    p4[1]= -0.5*size;    p4[2]= -(b*p4[1] + d)/c;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1; 
  }  

  if( fabs(b)<EPSILON )
  {
    p1[0]= -0.5*size;    p1[1]= -0.5*size;    p1[2]= -(a*p1[0] + d)/c;
    p2[0]=  0.5*size;    p2[1]= -0.5*size;    p2[2]= -(a*p2[0] + d)/c;
    p3[0]=  0.5*size;    p3[1]=  0.5*size;    p3[2]= -(a*p3[0] + d)/c;
    p4[0]= -0.5*size;    p4[1]=  0.5*size;    p4[2]= -(a*p4[0] + d)/c;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1; 
  }  


  if( fabs(c)<EPSILON )
  { 
    p1[0]= -0.5*size;    p1[1]= -(a*p1[0] + d)/b;    p1[2]= -0.5*size;
    p2[0]=  0.5*size;    p2[1]= -(a*p2[0] + d)/b;    p2[2]= -0.5*size;
    p3[0]=  0.5*size;    p3[1]= -(a*p3[0] + d)/b;    p3[2]=  0.5*size;
    p4[0]= -0.5*size;    p4[1]= -(a*p4[0] + d)/b;    p4[2]=  0.5*size;
    glBegin(GL_QUADS);
      glVertex3fv(p1);
      glVertex3fv(p2);
      glVertex3fv(p3);
      glVertex3fv(p4);
    glEnd();
    return 1; 
  }  


  p1[0]= -0.5*size;    p1[1]= -0.5*size;    p1[2]= -(a*p1[0] + b*p1[1] + d)/c;
  p2[0]=  0.5*size;    p2[1]= -0.5*size;    p2[2]= -(a*p2[0] + b*p2[1] + d)/c;
  p3[0]=  0.5*size;    p3[1]=  0.5*size;    p3[2]= -(a*p3[0] + b*p3[1] + d)/c;
  p4[0]= -0.5*size;    p4[1]=  0.5*size;    p4[2]= -(a*p4[0] + b*p4[1] + d)/c;

  glBegin(GL_QUADS);
    glVertex3fv(p1);
    glVertex3fv(p2);
    glVertex3fv(p3);
    glVertex3fv(p4);
  glEnd();  
  

  return 1;
}
#endif

//! Determines if a 2D point is inside a triangle or not.
//! \param p the point to test the inclusion
//! \param a first point of the triangle
//! \param b second point of the triangle
//! \param c third point of the triangle
//! \return true if the point is inside the triangle, false otherwise
bool isPointInTriangle(Vector2D p, Vector2D a, Vector2D b, Vector2D c)
{
  Vector2D u, v, w;

  u= b - a;
  v= p - a;
  w= c - a;
  if( ( u.y*v.x - u.x*v.y > 0 )  !=  ( u.y*w.x - u.x*w.y > 0 ) )
  {  return false;  }

  u= c - b;
  v= p - b;
  w= a - b;
  if( ( u.y*v.x - u.x*v.y > 0 )  !=  ( u.y*w.x - u.x*w.y > 0 ) )
  {  return false;  }

  u= a - c;
  v= p - c;
  w= b - c;
  if( ( u.y*v.x - u.x*v.y > 0 )  !=  ( u.y*w.x - u.x*w.y > 0 ) )
  {  return false;  }

  return true;
}

//! Computes the area of a polygon.
//! It uses the formula:
//! A= 0.5 * (x1*y2 - y1*x2  +  x2*y3 - y2*x3  +  ...  +  xn*y1 - yn*x1 )
//! The computed area is positive if the vertices are given in counter-clockwise order,
//! negative otherwise.
//! Weisstein, Eric W. "Polygon Area." From MathWorld--A Wolfram Web Resource.
//! http://mathworld.wolfram.com/PolygonArea.html
//! \param vertices a std:vector containing the vertices
//! \return the computed area (positive or negative according to the polygon orientation)
float polygonArea(const std::vector<Vector2D> &vertices)
{
  unsigned int i;
  float area= 0.0;

  for(i=0; i<vertices.size()-1; ++i)
  {
    area+= vertices[i].x*vertices[i+1].y - vertices[i].y*vertices[i+1].x;
  }
  area+= vertices[vertices.size()-1].x*vertices[0].y - vertices[vertices.size()-1].y*vertices[0].x;
  area*= 0.5;

  return area;
}

//! Computes the area of a polygon.
//! It uses the formula:
//! A= 0.5 * (x1*y2 - y1*x2  +  x2*y3 - y2*x3  +  ...  +  xn*y1 - yn*x1 )
//! The computed area is positive if the vertices are given in counter-clockwise order,
//! negative otherwise.
//! Weisstein, Eric W. "Polygon Area." From MathWorld--A Wolfram Web Resource.
//! http://mathworld.wolfram.com/PolygonArea.html
//! \param vertices a std:list containing the vertices
//! \return the computed area (positive or negative according to the polygon orientation)
float polygonArea(std::list<Vector2D> &vertices)
{
  float area= 0.0;
  std::list<Vector2D>::iterator current, next;
 
  current= vertices.begin();
  while(current!=vertices.end())
  {
     next= current;
     next++;
     if(next==vertices.end())
     {  next= vertices.begin();  }

     area+= (*current).x*(*next).y - (*current).y*(*next).x;

     current++;
  }

  area*= 0.5;

  return area;
}



//! Computes a quality score for a 2D triangle.
//! This score gives an idea of how close to equilateral is the triangle (the highest the closest).
//! \param p1 first vertex of the triangle
//! \param p2 second vertex of the triangle
//! \param p3 third vertex of the triangle
//! \return the computed quality
float triangleQuality(Vector2D p1, Vector2D p2, Vector2D p3)
{
  float a, b, c, s, area;
  Vector2D p1p2, p2p3, p3p1;

  p1p2= p2 - p1;
  p2p3= p3 - p2;
  p3p1= p1 - p3;


  //compute the lengths of the triangle edges:
  a= p1p2.norm();
  b= p2p3.norm();
  c= p3p1.norm();

  //compute the triangle area:
  s= ( a + b + c )/2;

  area= sqrt( s*(s-a)*(s-b)*(s-c) );  //triangle area

  // quality= area / sum of the squares of the edge lengths
  return area/(a*a + b*b + c*c); 
}


//! Computes the angle at the second vertex of the given triangle.
//! \param p1 first vertex of the triangle
//! \param p2 second vertex of the triangle
//! \param p3 third vertex of the triangle
//! \return the angle between p2p1 and p2p3
float triangleVertexAngle(Vector2D p1, Vector2D p2, Vector2D p3)
{
  float angle;
  Vector2D p2p1, p2p3;

  p2p1= p1 - p2;
  p2p3= p3 - p2;

  p2p1.normalize();
  p2p3.normalize();

  angle= fabs( acos( dotProduct(p2p1, p2p3) ) );

  return angle;
}

//! Triangulates a polygon. The function is based on the "ear cut" algorithm.
//! At each iteration, the choice of the ear to cut is not arbitrary but based on a criterion (see code below).
//! This choice can give a better triangulation.
//! The input polygon can have an arbitrary orientation (clockwise or counter clockwise).
//! \param vertex_array array of the vertex coordinates (x,y) of the polygon to triangulate 
//! \param triangles the computed triangle (a set of three indices in the vertex array) array
//! \return 1 in case of success, 0 otherwise
int triangulatePolygon(std::vector<Vector2D> &vertex_array, std::vector<Triangle> &triangles)
{
  bool ear_found;
  unsigned int i, nb_iterations;
  Vector2D p1, p2;
  Triangle triangle;
  float area, angle, min_angle= 1e10;
  std::list<unsigned int> vertex_list;
  std::list<unsigned int>::iterator previous, current, next, current2;
  std::list<unsigned int>::iterator previous_best, best, next_best;
  std::list<Triangle> triangle_list;
  std::list<Triangle>::iterator itriangle;  


  area= polygonArea(vertex_array);

  // fill the list with the vertex indices:
  for(i=0; i<vertex_array.size(); ++i)
  {   
    if( sign(area)==1 ) //CCW orientation
    {  vertex_list.push_back(i);  }
    else
    {  vertex_list.push_front(i);  }
  }

  nb_iterations= 0;
  while(true)
  {
    if(nb_iterations > vertex_array.size()-2)
    {
       printf("%s: %d: triangulatePolygon(): the number of iterations is abnormally high. Check the consistency of input data.\n", __FILE__, __LINE__);
       return 0; 
    }
    nb_iterations++;

    if(vertex_list.size()==3)
    { 
       previous= vertex_list.begin();
       current= previous;
       current++;
       next= current;
       next++;
       triangle.v[0]= *previous;
       triangle.v[1]= *current;
       triangle.v[2]= *next;
       triangle_list.push_back(triangle);
       vertex_list.clear();

       //convert the list to a vector:
       triangles.resize(triangle_list.size());
       i= 0;
       for(itriangle=triangle_list.begin(); itriangle!=triangle_list.end(); itriangle++)
       {
         triangles[i]= *itriangle;
         ++i;
       }
       triangle_list.clear();
      return 1;
    }
  
    ear_found= false;
    for(current=vertex_list.begin(); current!=vertex_list.end(); current++)
    {
       if(current==vertex_list.begin())
       {  previous= vertex_list.end();  }
       else 
       {  previous= current;  }
       previous--;

       next= current;
       next++;
       if(next==vertex_list.end())
       {  next= vertex_list.begin();  }

       p1= vertex_array[*current] - vertex_array[*previous];
       p2= vertex_array[*next] - vertex_array[*current];
       p1.normalize();
       p2.normalize();
     
       if( sign(p1.y*p2.x - p1.x*p2.y) == sign(area) ) // vertex is non convex
       { continue; }

       // check that all other vertices are outside the ear triangle:
       for(current2=vertex_list.begin(); current2!=vertex_list.end(); current2++)
       {  
          if(current2==previous || current2==current || current2==next)
          {  continue;  }

          if(isPointInTriangle(vertex_array[*current2], vertex_array[*previous], vertex_array[*current], vertex_array[*next]) )
          {  break;  }
       }

       // the vertex is an ear of the polygon:
       if(current2==vertex_list.end())
       {
         if(ear_found==false) // it is the first ear to be found
         {
           min_angle= triangleVertexAngle(vertex_array[*previous], vertex_array[*current], vertex_array[*next]);
           best= current;
           previous_best= previous;
           next_best= next;
           ear_found= true;
           continue;
         }
         else
         {
            angle= triangleVertexAngle(vertex_array[*previous], vertex_array[*current], vertex_array[*next]);
            if(angle < min_angle)
            {
              min_angle= angle;
              best= current;
              previous_best= previous;
              next_best= next;
              continue;
            }
         }
       }

    }

    if(ear_found)
    {    
      // add the new triangle:
      triangle.v[0]= *previous_best;
      triangle.v[1]= *best;
      triangle.v[2]= *next_best;    
  
      triangle_list.push_back(triangle);
  
      //remove the ear vertex from the list:
      vertex_list.erase(best);
    }
    else
    {
       printf("%s: %d: triangulatePolygon(): no \"ear\" vertex was found for the polygon. Check the consistency of input data.\n", __FILE__, __LINE__);
       return 0;
    }
  }


  return 0;
}

//! Builds a revolution surface (as a triangular mesh) from a 2D curve.
//! The curve will be rotated around its y-axis All the x coordinates of its points must have the same sign.
//! The first and last points of the curve do not need to have a null x coordinate.
//! If two points of the curve are very close to each other they will be merged.
//! \param curve the curve: a list of (x,y) points (the function accepts clockwise and counter-clockwise orientation)
//! \param alpha_min minimum rotation angle (in radians)
//! \param alpha_max maximum rotation angle (in radians). alpha_max must be > alpha_min
//! \param dalpha rotation angle step used to compute all the intermediate rotations (must be < alpha_max - alpha_min)
//! \param epsilon if the distance between two points of the curve is smaller than this value, they will be merged. This avoids to have excessively small triangles in the computed mesh.
//! \param mesh the mesh that will be filled with the computed surface
//! \return 1 in case of success, 0 otherwise
int surfaceOfRevolution(std::list<Vector2D> curve, float alpha_min, float alpha_max, float dalpha, float epsilon, TriangleMesh &mesh)
{
   bool duplicateFirst, duplicateLast;
   unsigned int i, j, nbRotations, shift, firsts_index= 0, lasts_index= 0;
   int Xsign;
   float alpha;
   Vector3D p;
   Vector2D first, last;
   Triangle triangle;
   std::vector<Vector2D> curve_array;
   std::vector<Triangle> triangles;
   std::list<Vector2D>::iterator current, next;


   if(alpha_min >= alpha_max)
   { 
     printf("%s: %d: surfaceOfRevolution(): alpha_min must be < alpha_max.\n",__FILE__,__LINE__); 
     return 0; 
   }


   if(curve.size() < 2)
   { 
     printf("%s: %d: surfaceOfRevolution(): the curve must have more than two vertices.\n",__FILE__,__LINE__); 
     return 0; 
   }

   current= curve.begin();
   while(current!=curve.end())
   {
     next= current;
     next++;
     if(next==curve.end())
     {  break;  }

     if( norm( (*current) - (*next) ) < 3*epsilon )
     {
       current= curve.erase(current);
       continue;
     } 

     current++;
   }
  
   if(curve.size() < 2)
   { 
     printf("%s: %d: surfaceOfRevolution(): there are less than three points after merging the very close points.\n",__FILE__,__LINE__); 
     return 0; 
   }


   Xsign= 0;
   for(current=curve.begin(); current!=curve.end(); current++)
   {
     if( fabs(current->x) > epsilon)
     {
        if(Xsign==0) //it is the first non null x coordinate
        {
           Xsign= sign(current->x);
        }
        else
        {
           if( sign(current->x)!=Xsign )
           {
             printf("%s: %d: surfaceOfRevolution(): all the vertex x coordinates must have the same sign.\n",__FILE__,__LINE__); 
             return 0;
           }
        }
     }
   }



   // change the order of the curve elements in case it is not in direct (counter-clockwise) orientation:
   if(polygonArea(curve)<0)
   { 
     curve.reverse();
   }

   first= curve.front();
   last = curve.back();


   
   //If the first point is almost not displaced by the rotations, it will not be duplicated.
   if( first.x*fabs(alpha_max-alpha_min) < epsilon )
   {  duplicateFirst= false;  }
   else
   {  duplicateFirst= true;  }

   //If the last point is almost not displaced by the rotations, it will not be duplicated.
   if( last.x*fabs(alpha_max-alpha_min) < epsilon)
   {  duplicateLast= false;  }
   else
   {  duplicateLast= true;  }
   

   if(!duplicateFirst)
   {  curve.pop_front();  }
   if(!duplicateLast)
   {  curve.pop_back();   }


   mesh.reset();
   
   nbRotations= (unsigned int) ( (alpha_max - alpha_min)/dalpha );
   dalpha= (alpha_max - alpha_min)/((float) nbRotations);

   for(i=0; i<nbRotations; ++i)
   {
     for(current=curve.begin(); current!=curve.end(); current++)
     {
        alpha= alpha_min + i*dalpha;
        p.x= cos(alpha)*current->x;
        p.y= (*current).y;
        p.z= sin(alpha)*current->x;
        mesh.addVertex(p.x, p.y, p.z);
     }
   }

   shift= curve.size();
   for(i=0; i<nbRotations-1; ++i)
   {
     for(j=0; j<curve.size()-1; j++)
     {
        triangle.setVertexIndices(j + i*shift, j + 1  + i*shift, j + (i+1)*shift);
        mesh.addTriangle(triangle);
        triangle.setVertexIndices(j + 1 + i*shift, j + 1 + (i+1)*shift, j + (i+1)*shift);
        mesh.addTriangle(triangle);
     }
   }

   
   if(!duplicateFirst)
   {
     mesh.addVertex(first.x, first.y, 0.0);
     firsts_index= mesh.nbVertices()-1;
   }
   if(!duplicateLast)
   {
     mesh.addVertex(last.x, last.y, 0.0);
     lasts_index= mesh.nbVertices()-1;
   }   


   // connect the first and last points to the structure:
   if(!duplicateFirst)
   { 
     for(i=0; i<nbRotations-1; ++i)
     {
       triangle.setVertexIndices(firsts_index, i*shift, (i+1)*shift);
       mesh.addTriangle(triangle);
     }
   }
   if(!duplicateLast)
   { 
     for(i=0; i<nbRotations-1; ++i)
     {
       triangle.setVertexIndices(lasts_index, shift-1  + (i+1)*shift, shift-1  + i*shift);
       mesh.addTriangle(triangle);
     }
   }

   // close the loop for each rotated curve:
   if(duplicateFirst)
   {  
      if(duplicateLast)
      {  
        for(i=0; i<nbRotations-1; ++i)
        {
          triangle.setVertexIndices(i*shift, (i+1)*shift, shift-1  + i*shift);
          mesh.addTriangle(triangle);
          triangle.setVertexIndices(shift-1 + i*shift, (i+1)*shift, shift-1 + (i+1)*shift);
          mesh.addTriangle(triangle);
        } 
      }
      else
      {
        for(i=0; i<nbRotations-1; ++i)
        {
          triangle.setVertexIndices(lasts_index, i*shift, (i+1)*shift);
          mesh.addTriangle(triangle);
        } 
      }
   }
   else
   {
      if(duplicateLast)
      {  
        for(i=0; i<nbRotations-1; ++i)
        {
          triangle.setVertexIndices(mesh.nbVertices()-1, shift-1 + (i+1)*shift, shift-1 + i*shift);
          mesh.addTriangle(triangle);
        } 
      }
   }

   
   curve_array.resize(curve.size());

   i= 0;
   for(current=curve.begin(); current!=curve.end(); current++)
   {
     curve_array[i]= (*current);
     ++i; 
   }


   // triangles on the two sides:
   triangulatePolygon(curve_array, triangles);

   for(i=0; i<triangles.size(); ++i)
   {
     triangle.setVertexIndices(triangles[i].v[0], triangles[i].v[2], triangles[i].v[1]);
     mesh.addTriangle(triangle);
     triangle.setVertexIndices(triangles[i].v[0]+(nbRotations-1)*shift, triangles[i].v[1]+(nbRotations-1)*shift, triangles[i].v[2]+(nbRotations-1)*shift);
     mesh.addTriangle(triangle);
   }   

   if(duplicateFirst)
   {
     if(!duplicateLast)
     {
       triangle.setVertexIndices(lasts_index, shift-1, 0);
       mesh.addTriangle(triangle);
       triangle.setVertexIndices(lasts_index, (nbRotations-1)*shift, (nbRotations-1)*shift + shift-1);
       mesh.addTriangle(triangle);
     }
   }
   else
   {
     if(duplicateLast)
     {
       triangle.setVertexIndices(firsts_index, shift-1, 0);
       mesh.addTriangle(triangle);
       triangle.setVertexIndices(firsts_index, (nbRotations-1)*shift, (nbRotations-1)*shift + shift-1);
       mesh.addTriangle(triangle);
     }
     else
     { 
       triangle.setVertexIndices(firsts_index, shift-1, 0);
       mesh.addTriangle(triangle);
       triangle.setVertexIndices(firsts_index, lasts_index, shift-1);
       mesh.addTriangle(triangle);
       triangle.setVertexIndices(lasts_index, (nbRotations-1)*shift, (nbRotations-1)*shift + shift-1);
       mesh.addTriangle(triangle);
       triangle.setVertexIndices(lasts_index, firsts_index, (nbRotations-1)*shift);
       mesh.addTriangle(triangle);
     }
   }

   //if x coordinates were < 0, invert the triangle normals: 
   if(Xsign==-1)
   {
     mesh.invertTriangleNormals();
   }

   return 1;
}


//! Extracts the main horizontal plane of a point set supposed to contain points on such a plane.
//! The function builds a histogram of the point heights (z-coordinate) and finds the bin
//! with the biggest value binMax.
//! Starting from this bin and going in increasing height, it selects all the bins until there is one
//! with a value less than binMax/alpha where alpha is a user-chosen parameter.
//! The height interval of the selected bins is then supposed to contain the plane points.
//! \param points the input set of points
//! \param quantizationStep the height interval of each bin
//! \param alpha parameter of selection of the bins (values between 2 and 5 give good results)
//! \param zmin the minimum height of the plane points
//! \param zmax the maximum height of the plane points
//! \return OK in case of success, ERROR otherwise
int extractHorizontalPlane(const std::vector<Vector3D> &points, float quantizationStep, float alpha, float &zmin, float &zmax)
{
  unsigned int i;
  float minZ, maxZ;
  std::vector< std::pair<float,int> > bins; // bin= [ height, number of points at this height ]
  int binMin, binMax, ibinMax;

  // get extreme values of z:
  minZ= maxZ= 0.0;
  for(i=0; i<points.size(); ++i)
  {
    if(i==0)
    {   minZ= maxZ= points[i].z;  }
    else
    {
      if( points[i].z < minZ )
      {  minZ= points[i].z;  }
      if( points[i].z > maxZ )
      {  maxZ= points[i].z;  }
    }
  }


  bins.assign( (unsigned int)  ((maxZ - minZ)/quantizationStep) + 1, std::make_pair(0,0) );

  // fills bins abscissae:
  for(i=0; i<bins.size(); ++i)
  {
    bins[i].first= minZ + i*quantizationStep;
  }

  // fills bins value:
  for(i=0; i<points.size(); ++i)
  {
    bins.at( (unsigned int) floor( (points[i].z - minZ)/quantizationStep ) ).second++;
  }

  // sort in increasing abscissa (point height):
  std::sort( bins.begin(), bins.end() ); 

  // find bin with the biggest number of elements:
  binMax= 0;
  ibinMax= 0;
  for(i=0; i<bins.size(); ++i)
  {
    if( i==0 || (bins[i].second > binMax) )
    {  
      binMax= bins[i].second;  
      ibinMax= i;
    }
  }

  zmin= bins[ibinMax].first;
  binMin= (int) ((float) binMax)/alpha;

  zmax= zmin;
  for(i=ibinMax+1; i<bins.size(); ++i)
  {
    if(bins[i].second < binMin)
    {  break;  }
    else
    {  zmax= bins[i].first; }
  }

  
  return OK;
}

//! Default constructor of OccupancyGrid.
OccupancyGrid::OccupancyGrid()
{
  position_.set(0.0, 0.0, 0.0);
  size_= 0.0;
  dimX_= dimY_= dimZ_= 0;
  zmin_= zmax_= 0.0;
  nbConnectedComponents_= 0;
}

//! Resets the content of the OccupancyGrid.
//! \return OK in case of success, ERROR otherwise
int OccupancyGrid::reset()
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::reset(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  grid_.clear();
  points_.clear();
  position_.set(0.0, 0.0, 0.0);
  size_= 0.0;
  dimX_= dimY_= dimZ_= 0;
  zmin_= zmax_= 0.0;
  nbConnectedComponents_= 0;

  return OK;
}


//! Gives access to an element of the grid from its three coordinate indices (each starting from 0).
//! \param i coordinate along X-axis in the grid array (starting from 0)
//! \param j coordinate along Y-axis in the grid array (starting from 0)
//! \param k coordinate along Z-axis in the grid array (starting from 0)
//! \return the content of the element at position (i,j,k), 0 in case of error
int OccupancyGrid::operator () (const unsigned int &i, const unsigned int &j, const unsigned int &k) const
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::operator(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return 0;
  }

  if( (i > dimX_-1) ||(j > dimY_-1) || (k > dimZ_-1) )
  {
    printf("%s: %d: OccupancyGrid::operator (): indices (%u,%u,%u) exceed grid dimensions (%u,%u,%u).\n",__FILE__,__LINE__,i,j,k,dimX_,dimY_,dimZ_);
    return 0;
  }

  return grid_.at(i + j*dimX_ + k*dimX_*dimY_);
}

//! Gives access to an element of the grid from its coordinate indices (starting from 0).
//! \param i coordinate along X-axis in the grid array (starting from 0)
//! \param j coordinate along Y-axis in the grid array (starting from 0)
//! \param k coordinate along Z-axis in the grid array (starting from 0)
//! \return  a reference to the element at position (i,j,k)
int& OccupancyGrid::operator () (const unsigned int &i, const unsigned int &j, const unsigned int &k)
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::operator(): the calling instance is NULL.\n",__FILE__,__LINE__);
  }

  if( (i > dimX_-1) || (j > dimY_-1) || (k > dimZ_-1) )
  {
    printf("%s: %d: OccupancyGrid::operator (): indices (%u,%u,%u) exceed grid dimensions (%u,%u,%u).\n",__FILE__,__LINE__,i,j,k,dimX_,dimY_,dimZ_);
    return grid_.at(0);
  }

  return grid_.at(i + j*dimX_ + k*dimX_*dimY_);
}

//! Finds the neighbors of a cell of the grid.
//! \param index index of the element in the cell array of the grid
//! \param neighborList list of the indices of the input-cell's neighbors. Uses 26-connectivity.
//! \return OK in case of success, ERROR otherwise
int OccupancyGrid::neighbors(const unsigned int &index, std::list<unsigned int> &neighborList)
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::neighbors(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  if( index > dimX_*dimY_*dimZ_-1 )
  {
    printf("%s: %d: OccupancyGrid::neighbors (): index (%u) exceed grid size (%u).\n",__FILE__, __LINE__, index, dimX_*dimY_*dimZ_);
    return ERROR;
  }

  int i, j, k, tmp;
  int i2, j2, k2;

  k= (index)/(dimX_*dimY_);
  tmp= index - k*dimX_*dimY_;
  j= tmp/dimX_;
  i= tmp - j*dimX_;

  neighborList.clear();
  for(i2=i-1; i2<i+2; ++i2)
  {
    if( (i2<0) || (i2>=(int)dimX_) )
    {  continue;  }

    for(j2=j-1; j2<j+2; ++j2)
    {
      if( (j2<0) || (j2>=(int)dimY_) )
      {  continue;  }

      for(k2=k-1; k2<k+2; ++k2)
      {
        if( (k2<0) || (k2>=(int)dimZ_) )
        {  continue;  }

        if( (i2==i) && (j2==j) && (k2==k) )
        {  continue;  }

//         if( (i2!=i) && (j2!=j) && (k2!=k) )
//         {  continue;  }
// 
//         if( (i2!=i && j2!=j) || (i2!=i && k2!=k) || (j2!=j && k2!=k) )
//         {  continue;  }
        neighborList.push_back( (unsigned int)(i2 + j2*dimX_ + k2*dimX_*dimY_) );
      }
    }
  }

  return OK;
}

//! Loads the point cloud contained in a text file.
//! Each line of the file must contains only the three coordinates of a point and
//! possibly comment after a '#'.
//! \param filename name of the file to load
//! \param delta edge length of the grid cells
//! \return OK in case of success, ERROR otherwise
int OccupancyGrid::loadFile(const std::string &filename, float delta)
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::loadFile(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  bool result;
  size_t pos;
  unsigned int i, j, k;
  int line_count;
  float x, y, z;
  float minX, maxX, minY, maxY, minZ, maxZ, zmin, zmax;
  std::string line;
  std::ifstream in;
  std::istringstream iss;
  Vector3D point;
  std::list<Vector3D> points;
  std::list<Vector3D>::iterator iter;
  Vector2D v;
  std::vector<Vector2D> plane;
  std::list<unsigned int> hull;

  in.open(filename.c_str());  
  if( in.fail() ) 
  {
    printf("%s: %d: OccupancyGrid::loadOBJfile(): can not open \"%s\".\n", __FILE__, __LINE__, filename.c_str());
    in.close();
    return ERROR;
  }

  while(in.good())
  {
    getline(in, line);
    line_count++;

     //Remove all that's after a '#' character
     pos= line.find_first_of('#');
     if(pos!=line.npos)
     { 
       if(pos==0)
       {  continue; }
       line= line.substr(0, pos);
     }

     //Remove all the spaces at the end of the line:
     pos= line.find_last_not_of(' ');
     line= line.substr(0, pos+1);

     if( line.empty() )
     {  continue;  }

     //Create a stream from the line string:
     iss.clear();
     iss.str(line);

     result= ( iss >> x >> y >> z );

     point.set(x, y, z);
     points.push_back(point);
  }
  in.close();

  minX= maxX= minY= maxY= minZ= maxZ= 0.0;
  for(iter=points.begin(); iter!=points.end(); ++iter)
  {
    if(iter==points.begin())
    {
      minX= maxX= iter->x;
      minY= maxY= iter->y;
      minZ= maxZ= iter->z;
    }
    else
    {
      if( iter->x < minX )
      {  minX= iter->x;  }
      if( iter->x > maxX )
      {  maxX= iter->x;  }

      if( iter->y < minY )
      {  minY= iter->y;  }
      if( iter->y > maxY )
      {  maxY= iter->y;  }

      if( iter->z < minZ )
      {  minZ= iter->z;  }
      if( iter->z > maxZ )
      {  maxZ= iter->z;  }
    }
  }

  position_.set(minX, minY, minZ);

  size_= delta;

  dimX_= (unsigned int) ( (maxX - minX)/delta ) + 1;
  dimY_= (unsigned int) ( (maxY - minY)/delta ) + 1;
  dimZ_= (unsigned int) ( (maxZ - minZ)/delta ) + 1;

  if( (dimX_*dimY_*dimZ_) > 1500000 )
  {
    printf("%s: %d: OccupancyGrid::loadOBJfile(): grid size would be too big (%d).\n", __FILE__, __LINE__,dimX_*dimY_*dimZ_);
    reset();
    return ERROR;
  }

  points_.resize(points.size());
  std::copy(points.begin(), points.end(), points_.begin());
  extractHorizontalPlane(points_, 0.0025, 3.0, zmin, zmax);

  grid_.assign(dimX_*dimY_*dimZ_, 0);

  // mark the cell as occupied if they contain points with height > zmax:
  for(iter=points.begin(); iter!=points.end(); ++iter)
  {
    if( iter->z < zmax )
    {  continue;  }

    i= floor( (iter->x - minX)/size_ );
    j= floor( (iter->y - minY)/size_ );
    k= floor( (iter->z - minZ)/size_ );

    (*this)(i, j, k)= 1;
  }

  // get the points of the table ( zmin < z < zmax):
  for(iter=points.begin(); iter!=points.end(); ++iter)
  {
    if( (iter->z > zmin) && (iter->z < zmax) )
    { 
      v.set( iter->x, iter->y );
      plane.push_back(v);
    }
  }

  zmin_= zmin;
  zmax_= zmax;

  if(plane.size() >= 3)
  {
    // only keep the table point that are on the convex hull:
    convexHull2D(plane, hull);
  
    for(std::list<unsigned int>::iterator ihull=hull.begin(); ihull!=hull.end(); ++ihull)
    {
      plane_.push_back( plane.at(*ihull) );
    }
  }


  return OK;
}

#if defined(ENABLE_DISPLAY)
//! Draws the grid cells and the original point cloud.
//! Cells are displayed with different colors according to their label.
//! \return OK in case of success, ERROR otherwise
int OccupancyGrid::draw()
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::draw(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  unsigned int i, j, k;
  int value;
  //GLfloat color[3];
  Vector3D center;

  glPushAttrib(GL_ENABLE_BIT | GL_POINT_BIT | GL_LINE_BIT);
  glEnable(GL_CULL_FACE);

  glPointSize(3);
  glDisable(GL_LIGHTING);
  glColor3f(1, 0, 1);
  glBegin(GL_POINTS);
    for(i=0; i<points_.size(); ++i)
    glVertex3f(points_[i].x, points_[i].y, points_[i].z);
  glEnd();


  glLineWidth(3);
  glColor3f(0, 0, 0);
  glBegin(GL_LINE_LOOP);
    for(i=0; i<plane_.size(); ++i)
    {  glVertex3f(plane_[i].x, plane_[i].y, zmin_);  }
  glEnd();

  glBegin(GL_LINE_LOOP);
    for(i=0; i<plane_.size(); ++i)
    {  glVertex3f(plane_[i].x, plane_[i].y, zmax_);  }
  glEnd();

  glBegin(GL_LINE_STRIP);
    for(i=0; i<plane_.size(); ++i)
    {
      glVertex3f(plane_[i].x, plane_[i].y, zmax_);
      glVertex3f(plane_[i].x, plane_[i].y, zmin_);
    } 
  glEnd();
  glLineWidth(1);
  glEnable(GL_LIGHTING);

  glPushMatrix();
    glTranslatef(position_.x, position_.y, position_.z);
    glColor3f(0, 1, 0); 
      for(i=0; i<dimX_; ++i)
      {
        for(j=0; j<dimY_; ++j)
        {
          for(k=0; k<dimZ_; ++k)
          {
            value= (*this)(i, j, k);
  
            if(value==0)
            {  continue;  }
            if(value < 0)
            {   glColor3f(1, 0, 0);   } 
            else
            {
              glColor3f(0, 1, 1);
            }

            center.set( (i+0.5)*size_, (j+0.5)*size_, (k+0.5)*size_ );
            drawCube(center, size_);
          }
        }
      }
    glPopMatrix();

  glPopAttrib();

  return OK;
}
#endif

//! Labels the grid cells by 3D connected components (clusters).
//! All the components that have less than a specified number of elementds (cells) are regarded as noise and removed.
//! \param minClusterSize all the clusters that have less elements than this value will be removed
//! \return OK in case of success, ERROR otherwise
int OccupancyGrid::labeling(unsigned int minClusterSize)
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::segment(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  unsigned int i, j, index, size;
  int label, label2;
  std::list<unsigned int> openList, neighborList;
 
  label= 2;
  label2= 2;
  std::map<int,int> labelMap;

  for(i=0; i<dimX_*dimY_*dimZ_; ++i)
  {
    if( grid_[i]!=1 )
    {  continue;  }

    label++;
    size= 0;
    openList.clear();

    openList.push_back(i);

    while( !openList.empty() )
    {
      index= openList.back();
      openList.pop_back();

      if(grid_.at(index)!=1)
      {  continue;  }

      grid_.at(index)= label;
      size++;
      neighbors(index, neighborList);
      openList.splice(openList.end(), neighborList);
    }

    if(size < minClusterSize)
    {
      for(j=0; j<dimX_*dimY_*dimZ_; ++j)
      {
        if( grid_[j]==label )
        {  grid_[j]= -1;  }
      }
    }
    else
    {
      labelMap[label]= label2;
      label2++;
    }
  }

  nbConnectedComponents_= label2 - 2;

  for(i=0; i<dimX_*dimY_*dimZ_; ++i)
  {
    if( grid_[i] < 2 )
    {  continue;  }
    else
    {
      grid_[i]= labelMap[grid_[i]];
    }
  }

  return OK;
}
#if defined(ENABLE_DISPLAY)
//! Exports the grid content as a vector of TriangleMesh in order to create bodies to be used for collision detection.
//! \param meshes the output vector of TriangleMesh
//! \return OK in case of success, ERROR otherwise
int  OccupancyGrid::exportAsTriangleMeshes(std::vector<TriangleMesh> &meshes)
{
  if(this==NULL)
  {
    printf("%s: %d: OccupancyGrid::exportAsTriangleMeshes(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  int i, j, k;
  int label;
  std::list<unsigned int> neighborList;
  std::map<int,unsigned int> labelMap; // (label), (index of a mesh int the vector of meshes)
  std::map<int,unsigned int>::iterator iter;
  Vector3D a, b, c, d, centroidTop, centroidBottom;
  TriangleMesh mesh;

  meshes.reserve(nbConnectedComponents_);

  //convert the plane points:
  if(plane_.size() >= 2)
  {
    centroidTop.set(0.0, 0.0, 0.0);

    mesh.reset();
    meshes.push_back(mesh);

    // side triangles:
    for(i=0; i<(int)(plane_.size()); ++i)
    {  
      a.set(plane_[i].x, plane_[i].y, zmax_);
      b.set(plane_[i].x, plane_[i].y, zmin_);
      if( i != (int)(plane_.size()-1) )
      {
        c.set(plane_[i+1].x, plane_[i+1].y, zmax_);
        d.set(plane_[i+1].x, plane_[i+1].y, zmin_);
      }
      else
      {
        c.set(plane_[0].x, plane_[0].y, zmax_);
        d.set(plane_[0].x, plane_[0].y, zmin_);
      }
  
      glVertex3f(plane_[i].x, plane_[i].y, zmin_);
      meshes.back().addTriangle(a, b, d);
      meshes.back().addTriangle(d, c, a);
    
      centroidTop.x+= plane_[i].x;
      centroidTop.y+= plane_[i].y;
    }

    centroidTop/= (float) (plane_.size());
    centroidTop.z= zmax_;
    centroidBottom.set(centroidTop.x, centroidTop.y, zmin_);

    // top and bottom triangles:
    for(i=0; i<(int)(plane_.size()); ++i)
    {
      if( i != (int)(plane_.size()-1) )
      {
        b.set(plane_[i].x, plane_[i].y, zmax_);
        c.set(plane_[i+1].x, plane_[i+1].y, zmax_);
        meshes.back().addTriangle(centroidTop, b, c);

        b.set(plane_[i].x, plane_[i].y, zmin_);
        c.set(plane_[i+1].x, plane_[i+1].y, zmin_);
        meshes.back().addTriangle(centroidBottom, c, b);
      }
      else
      {
        b.set(plane_[i].x, plane_[i].y, zmax_);
        c.set(plane_[0].x, plane_[0].y, zmax_);
        meshes.back().addTriangle(centroidTop, b, c);

        b.set(plane_[i].x, plane_[i].y, zmin_);
        c.set(plane_[0].x, plane_[0].y, zmin_);
        meshes.back().addTriangle(centroidBottom, c, b);
      }
    }
  }


  // convert each set of cubes with a same label into a triangle mesh (only keep the triangles of the envelope):
  for(i=0; i<((int)dimX_); ++i)
  {
    for(j=0; j<((int)dimY_); ++j)
    {
      for(k=0; k<((int)dimZ_); ++k)
      {
        label= (*this)(i,j,k);

        if(label < 1)
        {  continue;  }

        if(labelMap.count(label)==0)
        {   
          mesh.reset();
          meshes.push_back(mesh);
          labelMap[label]= meshes.size() - 1; 
        }

        if( (i-1)>=0 )
        { 
          if( (*this)(i-1,j,k)!=label )
          { 
            a.set(position_.x + i*size_, position_.y + j*size_, position_.z + k*size_);
            b.set(position_.x + i*size_, position_.y + j*size_, position_.z + (k+1)*size_);
            c.set(position_.x + i*size_, position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            d.set(position_.x + i*size_, position_.y + (j+1)*size_, position_.z + k*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }
        if( (i+1)<(int)(dimX_-1) )
        { 
          if( (*this)(i+1,j,k)!=label )
          { 
            a.set(position_.x + (i+1)*size_, position_.y + j*size_, position_.z + k*size_);
            b.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + k*size_);
            c.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            d.set(position_.x + (i+1)*size_, position_.y + j*size_, position_.z + (k+1)*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }


        if( (j-1)>=0 )
        { 
          if( (*this)(i,j-1,k)!=label )
          { 
            a.set(position_.x + i*size_    , position_.y + j*size_, position_.z + k*size_);
            b.set(position_.x + (i+1)*size_, position_.y + j*size_, position_.z + k*size_);
            c.set(position_.x + (i+1)*size_, position_.y + j*size_, position_.z + (k+1)*size_);
            d.set(position_.x + i*size_, position_.y + j*size_, position_.z + (k+1)*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }
        if( (j+1)<(int)(dimY_-1) )
        { 
          if( (*this)(i,j+1,k)!=label )
          { 
            a.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + k*size_);
            b.set(position_.x + i*size_    , position_.y + (j+1)*size_, position_.z + k*size_);
            c.set(position_.x + i*size_    , position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            d.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }


        if( (k-1)>=0 )
        { 
          if( (*this)(i,j,k-1)!=label )
          { 
            a.set(position_.x + (i+1)*size_, position_.y + j*size_    , position_.z + k*size_);
            b.set(position_.x + i*size_    , position_.y + j*size_    , position_.z + k*size_);
            c.set(position_.x + i*size_    , position_.y + (j+1)*size_, position_.z + k*size_);
            d.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + k*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }
        if( (k+1)<(int)(dimZ_-1) )
        { 
          if( (*this)(i,j,k+1)!=label )
          { 
            a.set(position_.x + i*size_    , position_.y + j*size_    , position_.z + (k+1)*size_);
            b.set(position_.x + (i+1)*size_, position_.y + j*size_    , position_.z + (k+1)*size_);
            c.set(position_.x + (i+1)*size_, position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            d.set(position_.x + i*size_    , position_.y + (j+1)*size_, position_.z + (k+1)*size_);
            meshes.at(labelMap[label]).addTriangle(a, b, c);
            meshes.at(labelMap[label]).addTriangle(c, d, a);
          }
        }

      }
    }
  }

  return OK;
}

#endif
//! Computes the convex hull of a set of 2D points.
//! It uses the incremental algorithm. 
//! \param points the input point set
//! \param result the convex hull (indices in the input array, in CCW order)
//! \return OK in case of success, ERROR otherwise
int convexHull2D(const std::vector<Vector2D> &points, std::list<unsigned int> &result)
{
  bool isInside;
  unsigned int tri1= 0, tri2= 1, tri3= 2;
  unsigned int i, pinsert;
  float cross, xmin= 0.0, xmax= 0.0;
  std::list<unsigned int> unprocessedPoints;
  std::list< std::pair< std::pair<unsigned int,unsigned int> , bool > > hull; // edge [index of vertex 1,index of vertex 2] + boolean [(]visible or not]
  std::list< std::pair< std::pair<unsigned int,unsigned int> , bool > >::iterator cur, prev, next, firstVisible, lastVisible;
  Vector2D a, b;


  hull.clear();

  // first find the starting hull as a triangle:

  // find the two extreme points along X-axis:
  for(i=0; i<points.size(); ++i)
  {
    if(i==0)
    {  xmin= xmax= points[i].x;  }
    else
    {
      if(points[i].x < xmin)
      {  
        tri1= i;
        xmin= points[i].x;
      }
      if(points[i].x > xmax)
      {  
        tri2= i;
        xmax= points[i].x;
      }
    }
  }
  // find a third point (not aligned on the two selected ones) to make a triangle
  for(i=0; i<points.size(); ++i)
  {
    if(i==tri1 || i==tri2)
    {  continue;  }
    else
    {
      a= points.at(tri2) - points.at(tri1);
      b= points.at(i) - points.at(tri1);
      
      cross= a.x*b.y - a.y*b.x;
      if( fabs(cross) < 10e-8 )
      {  continue;   }
      else
      {
        if(cross > 0)
        {
          tri3= i;
        }
        else
        {
          tri3= tri2;
          tri2= i;
        }
        break; 
      }
    }
  }
  if(i==points.size())
  {
    printf("%s: %d: convexHull2D(): could not find three non aligned points.\n",__FILE__,__LINE__);
    return ERROR;
  } 

  hull.push_back( std::make_pair( std::make_pair(tri1,tri2), false) );
  hull.push_back( std::make_pair( std::make_pair(tri2,tri3), false) );
  hull.push_back( std::make_pair( std::make_pair(tri3,tri1), false) );

  for(i=0; i<points.size(); ++i)
  {
    if(i==tri1 || i==tri2 || i==tri3)
    {   continue;   }
    unprocessedPoints.push_back(i);
  }

  while( !unprocessedPoints.empty() ) 
  {
    isInside= true;
    pinsert= unprocessedPoints.front();
    unprocessedPoints.pop_front();

    //compute the visibility from the current point for all the hull edges:
    for(cur=hull.begin(); cur!=hull.end(); ++cur)
    {
      a= points.at( cur->first.second ) - points.at( cur->first.first );
      b= points.at(pinsert) - points.at( cur->first.first );

      cross= a.x*b.y - a.y*b.x;

      if( cross < 0) //point visible 
      {
        cur->second= true;
        isInside= false;
      }
      else
      { cur->second= false; }
    }

//     printf("--hull= [");
//     for(cur2=hull.begin(); cur2!=hull.end(); ++cur2)
//     {   
//      if(cur2->second)
//       printf(" %d-%d", cur2->first.first, cur2->first.second);
//      else
//       printf(" %d %d", cur2->first.first, cur2->first.second);
//     }
//     printf("]\n");

    
    if(isInside) // point is inside the current hull
    {  continue;  }


    // look for the edges (there is one or two of them) that have only one adjacent edge visible:
    cur= hull.begin();
    while(cur!=hull.end())
    {
      if(cur->second==false) // edge is not visible from the point
      {  
         cur++;
         continue;
      }

      // get previous and next edges:
      if(cur==hull.begin())
      {  prev= (--hull.end())++;   }
      else
      {  prev= (--cur)++;     }

      next= (++cur)--; 
      if(next==hull.end())
      {  next= hull.begin();   }

      // both adjacent edge are visible
      if(prev->second==true && next->second==true)
      {
        cur++;
        continue;
      }

      // previous and next edges are not visible
      if(prev->second==false && next->second==false)
      { 
        firstVisible= cur;
        lastVisible = cur;
        cur++;
        break;
      }

      if(prev->second==false)
      {  
        firstVisible= cur;
        cur++;
        continue;
      }

      if(next->second==false)
      { 
        lastVisible= cur;
        cur++;
        continue;
      }


      printf("error %d [%d %d] [%d %d] [%d %d]\n",__LINE__,prev->first.first, prev->first.second,cur->first.first, cur->first.second,next->first.first, next->first.second);
      cur++;
      continue;
    }


    // replace the unique visible edge by two edges: 
    if(firstVisible==lastVisible)
    {
      next= (++firstVisible)--; 
      if(next==hull.end())
      {  next= hull.begin();   }

      firstVisible->first.second= pinsert;
      hull.insert(next, std::make_pair( std::make_pair(pinsert,next->first.first), false) );
    }
    else
    { 
      // connect the two extreme visible edges to each other:
      firstVisible->first.second = pinsert;
      lastVisible->first.first   = pinsert;
      firstVisible->second = false;
      lastVisible->second  = false;

      // delete all the edges that are still marked as visible:
      cur= hull.begin();
      while(cur!=hull.end())
      {
        if(cur->second==true)
        {
          cur= hull.erase(cur);
          continue;
        }
        cur++;
      }
    }

  }

  result.clear();

  for(cur=hull.begin(); cur!=hull.end(); ++cur)
  {
    result.push_back( cur->first.first );
  }

  return OK;
}

