#ifndef WORKSPACE_H
#define WORKSPACE_H

Vector3D SAHfingerForwardKinematics(float q1, float q2, float q3, float length1, float length2, float length3);

int SAHfingerWorkspace(float dq, float q1min, float q1max, float q2min, float q2max, float q3min, float q3max, float length1, float length2, float length3);

int linearRegression2D(std::vector<Vector2D> &points, Vector2D &result);

int linearRegression3D(std::vector<Vector3D> &points, Vector3D &result);

int bestFittingPlane(std::vector<Vector3D> &points, float result[4]);

int drawPlane(float a, float b, float c, float d, float size);

bool isPointInTriangle(Vector2D p, Vector2D a, Vector2D b, Vector2D c);

float polygonArea(const std::vector<Vector2D> &vertices);

float polygonArea(std::list<Vector2D> &vertices);

float triangleQuality(Vector2D p1, Vector2D p2, Vector2D p3);

float triangleVertexAngle(Vector2D p1, Vector2D p2, Vector2D p3);

int triangulatePolygon(std::vector<Vector2D> &vertex_array, std::vector<Triangle> &triangles);

int surfaceOfRevolution(std::list<Vector2D> curve, float alpha_min, float alpha_max, float dalpha, float epsilon, TriangleMesh &mesh);


int extractHorizontalPlane(const std::vector<Vector3D> &points, float quantizationStep, float alpha, float &zmin, float &zmax);
 
/*! \brief The OccupancyGrid class is used to convert a point cloud into data that can be exploited by the collision checker. */
class OccupancyGrid
{
  private:
   Vector3D position_; /*!< position of the (0,0,0) corner */
   float size_; /*!< edge length of the cells */
   unsigned int dimX_, dimY_, dimZ_; /*!< number of cells in each dimension */
   std::vector<int> grid_; /*!< cell array */
   std::vector<Vector3D> points_; /*!< original point cloud */
   std::vector<Vector2D> plane_; /*!< points of the hull of the plane points  */
   float zmin_, zmax_; /*!< lower and upper bounds on the height of the plane points  */
   unsigned int nbConnectedComponents_;  /*!< number of connected components of the grid (it is updated by a call to the labeling function) */

   int neighbors(const unsigned int &index, std::list<unsigned int> &neighborList);

  public:
   float size() { return size_; }
   unsigned int dimX() { return dimX_; }
   unsigned int dimY() { return dimY_; }
   unsigned int dimZ() { return dimZ_; }

   OccupancyGrid();
   int reset();
   int operator () (const unsigned int &i, const unsigned int &j, const unsigned int &k) const;
   int& operator () (const unsigned int &i, const unsigned int &j, const unsigned int &k);


   int loadFile(const std::string &filename, float delta);
   int draw();
//    int filter();
   int labeling(unsigned int minClusterSize);
   int exportAsTriangleMeshes(std::vector<TriangleMesh> &meshList);
};

int convexHull2D(const std::vector<Vector2D> &points, std::list<unsigned int> &result);
#endif
